#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
PDF’s standard security handler shall allow access permissions and up to two passwords to be specified for a
document: an owner password and a user password. An application’s decision to encrypt a document shall be
based on whether the user creating the document specifies any passwords or access restrictions.
"""
import hashlib
import typing
import zlib

from borb.io.read.encryption.rc4 import RC4
from borb.io.read.pdf_object import PDFObject
from borb.io.read.types import (
    AnyPDFType,
    Boolean,
    Decimal,
    HexadecimalString,
    Name,
    Reference,
    Stream,
    String,
    Dictionary,
)


class StandardSecurityHandler:
    """
    PDF’s standard security handler shall allow access permissions and up to two passwords to be specified for a
    document: an owner password and a user password. An application’s decision to encrypt a document shall be
    based on whether the user creating the document specifies any passwords or access restrictions.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        encryption_dictionary: Dictionary,
        user_password: typing.Optional[str] = None,
        owner_password: typing.Optional[str] = None,
    ):

        # (Optional) A code specifying the algorithm to be used in encrypting and
        # decrypting the document:
        # 0 An algorithm that is undocumented. This value shall not be used.
        # 1 "Algorithm 1: Encryption of data using the RC4 or AES algorithms" in 7.6.2,
        # "General Encryption Algorithm," with an encryption key length of 40 bits; see
        # below.
        # 2 (PDF 1.4) "Algorithm 1: Encryption of data using the RC4 or AES algorithms"
        # in 7.6.2, "General Encryption Algorithm," but permitting encryption key lengths
        # greater than 40 bits.
        # 3 (PDF 1.4) An unpublished algorithm that permits encryption key lengths
        # ranging from 40 to 128 bits. This value shall not appear in a conforming PDF
        # file.
        # 4 (PDF 1.5) The security handler defines the use of encryption and decryption
        # in the document, using the rules specified by the CF, StmF, and StrF entries.
        # The default value if this entry is omitted shall be 0, but when present should be a
        # value of 1 or greater.
        self._v = int(encryption_dictionary.get("V", Decimal(0)))

        # (Required) A 32-byte string, based on the user password, that shall be
        # used in determining whether to prompt the user for a password and, if so,
        # whether a valid user or owner password was entered. For more
        # information, see 7.6.3.4, "Password Algorithms."
        self._u: bytes = (
            StandardSecurityHandler._str_to_bytes(
                StandardSecurityHandler._unescape_pdf_syntax(
                    encryption_dictionary.get("U")
                )
            )
            or b""
        )
        assert len(self._u) == 32

        # (Required) A 32-byte string, based on both the owner and user passwords,
        # that shall be used in computing the encryption key and in determining
        # whether a valid owner password was entered. For more information, see
        # 7.6.3.3, "Encryption Key Algorithm," and 7.6.3.4, "Password Algorithms."
        self._o: bytes = (
            StandardSecurityHandler._str_to_bytes(
                StandardSecurityHandler._unescape_pdf_syntax(
                    encryption_dictionary.get("O")
                )
            )
            or b""
        )
        assert self._o is not None
        assert len(self._o) == 32

        # /ID
        trailer: typing.Optional[PDFObject] = encryption_dictionary.get_parent()
        assert trailer is not None
        assert isinstance(trailer, Dictionary)
        if "ID" in trailer:
            self._document_id: bytes = trailer["ID"][0].get_content_bytes()

        # (Required) A set of flags specifying which operations shall be permitted
        # when the document is opened with user access (see Table 22).
        assert "P" in encryption_dictionary
        self._permissions: int = int(encryption_dictionary.get("P"))  # type: ignore [arg-type]

        # (Optional; PDF 1.4; only if V is 2 or 3) The length of the encryption key, in bits.
        # The value shall be a multiple of 8, in the range 40 to 128. Default value: 40.
        # fmt: off
        self._key_length: int = int(encryption_dictionary.get("Length", Decimal(40)))
        assert self._key_length % 8 == 0, "The length of the encryption key, in bits must be a multiple of 8."
        # fmt: on

        # (Required) A number specifying which revision of the standard security
        # handler shall be used to interpret this dictionary
        self._revision: int = int(encryption_dictionary.get("R", Decimal(0)))

        # (Optional; meaningful only when the value of V is 4; PDF 1.5) Indicates
        # whether the document-level metadata stream (see 14.3.2, "Metadata
        # Streams") shall be encrypted. Conforming products should respect this
        # value.
        # Default value: true.
        self._encrypt_metadata: bool = encryption_dictionary.get(
            "EncryptMetadata", Boolean(True)
        )

        # verify password(s)
        password: typing.Optional[bytes] = None
        if user_password is not None:
            self._authenticate_user_password(bytes(user_password, encoding="charmap"))
            password = bytes(user_password, encoding="charmap")
        if owner_password is not None:
            self._authenticate_owner_password(bytes(owner_password, encoding="charmap"))
            password = bytes(owner_password, encoding="charmap")

        # calculate encryption_key
        # assert password is not None
        self._encryption_key: bytes = self._compute_encryption_key(password)

    #
    # PRIVATE
    #

    def _authenticate_owner_password(self, owner_password: bytes) -> bool:
        """
        Algorithm 7: Authenticating the owner password
        """
        # a) Compute an encryption key from the supplied password string, as described in steps (a) to (d) of
        # "Algorithm 3: Computing the encryption dictionary’s O (owner password) value".

        # b) (Security handlers of revision 2 only) Decrypt the value of the encryption dictionary’s O entry, using an RC4
        # encryption function with the encryption key computed in step (a).
        # (Security handlers of revision 3 or greater) Do the following 20 times: Decrypt the value of the encryption
        # dictionary’s O entry (first iteration) or the output from the previous iteration (all subsequent iterations),
        # using an RC4 encryption function with a different encryption key at each iteration. The key shall be
        # generated by taking the original key (obtained in step (a)) and performing an XOR (exclusive or) operation
        # between each byte of the key and the single-byte value of the iteration counter (from 19 to 0).

        # c) The result of step (b) purports to be the user password. Authenticate this user password using "Algorithm
        # 6: Authenticating the user password". If it is correct, the password supplied is the correct owner password.

        return False

    def _authenticate_user_password(self, user_password: bytes) -> bool:
        """
        Algorithm 6: Authenticating the user password
        """
        # a) Perform all but the last step of "Algorithm 4: Computing the encryption dictionary’s U (user password)
        # value (Security handlers of revision 2)" or "Algorithm 5: Computing the encryption dictionary’s U (user
        # password) value (Security handlers of revision 3 or greater)" using the supplied password string.
        previous_u_value: bytes = self._u
        self._compute_encryption_dictionary_u_value(user_password)
        u_value: bytes = self._u
        self._u = previous_u_value

        # b)If the result of step (a) is equal to the value of the encryption dictionary’s U entry (comparing on the first 16
        # bytes in the case of security handlers of revision 3 or greater), the password supplied is the correct user
        # password. The key obtained in step (a) (that is, in the first step of "Algorithm 4: Computing the encryption
        # dictionary’s U (user password) value (Security handlers of revision 2)" or "Algorithm 5: Computing the
        # encryption dictionary’s U (user password) value (Security handlers of revision 3 or greater)") shall be used
        # to decrypt the document.
        return self._u == u_value

    def _compute_encryption_dictionary_o_value(
        self,
        owner_password: typing.Optional[bytes],
        user_password: typing.Optional[bytes],
    ) -> bytes:
        """
        Algorithm 3: Computing the encryption dictionary’s O (owner password) value
        """

        # a) Pad or truncate the owner password string as described in step (a) of "Algorithm 2: Computing an
        # encryption key". If there is no owner password, use the user password instead.
        padded_owner_password: bytes = StandardSecurityHandler._pad_or_truncate(
            owner_password
        )

        # b) Initialize the MD5 hash function and pass the result of step (a) as input to this function.
        h = hashlib.md5()
        h.update(padded_owner_password)

        # c) (Security handlers of revision 3 or greater) Do the following 50 times: Take the output from the previous
        # MD5 hash and pass it as input into a new MD5 hash.
        if self._revision >= 3:
            prev_digest: bytes = h.digest()[0 : int(self._key_length / 8)]
            for _ in range(0, 50):
                h = hashlib.md5()
                h.update(prev_digest)
                prev_digest = h.digest()[0 : int(self._key_length / 8)]

        # d) Create an RC4 encryption key using the first n bytes of the output from the final MD5 hash, where n shall
        # always be 5 for security handlers of revision 2 but, for security handlers of revision 3 or greater, shall
        # depend on the value of the encryption dictionary’s Length entry.
        key: bytes = h.digest()[0:5]
        if self._revision >= 3:
            key = h.digest()[0 : int(self._key_length / 8)]

        # e) Pad or truncate the user password string as described in step (a) of "Algorithm 2: Computing an encryption
        # key".
        padded_user_password: bytes = StandardSecurityHandler._pad_or_truncate(
            user_password
        )

        # f) Encrypt the result of step (e), using an RC4 encryption function with the encryption key obtained in step
        # (d).
        rc4: RC4 = RC4()
        owner_key: bytes = rc4.encrypt(key, padded_user_password)

        # g) (Security handlers of revision 3 or greater) Do the following 19 times: Take the output from the previous
        # invocation of the RC4 function and pass it as input to a new invocation of the function; use an encryption
        # key generated by taking each byte of the encryption key obtained in step (d) and performing an XOR
        # (exclusive or) operation between that byte and the single-byte value of the iteration counter (from 1 to 19).
        if self._revision >= 3:
            for i in range(1, 20):
                key2: bytes = bytes([b ^ i for b in key])
                owner_key = RC4().encrypt(key2, owner_key)

        # h) Store the output from the final invocation of the RC4 function as the value of the O entry in the encryption
        # dictionary.
        self._o = owner_key
        return owner_key

    def _compute_encryption_dictionary_u_value(self, user_password_string: bytes):
        """
        Algorithm 4: Computing the encryption dictionary’s U (user password) value (Security handlers of revision 2)
        Algorithm 5: Computing the encryption dictionary’s U (user password) value (Security handlers of revision 3 or greater)
        """
        if self._revision == 2:
            # a) Create an encryption key based on the user password string, as described in "Algorithm 2: Computing an
            # encryption key".
            key_rev_2: bytes = self._compute_encryption_key(user_password_string)

            # b) Encrypt the 32-byte padding string shown in step (a) of "Algorithm 2: Computing an encryption key", using
            # an RC4 encryption function with the encryption key from the preceding step.

            # c) Store the result of step (b) as the value of the U entry in the encryption dictionary.
            self._u = RC4().encrypt(
                key_rev_2, StandardSecurityHandler._pad_or_truncate(None)
            )

            return self._u

        if self._revision >= 3:
            # a) Create an encryption key based on the user password string, as described in "Algorithm 2: Computing an
            # encryption key".
            key_rev_3: bytes = self._compute_encryption_key(user_password_string)

            # b) Initialize the MD5 hash function and pass the 32-byte padding string shown in step (a) of "Algorithm 2:
            # Computing an encryption key" as input to this function.
            h = hashlib.md5()
            h.update(StandardSecurityHandler._pad_or_truncate(None))

            # c) Pass the first element of the file’s file identifier array (the value of the ID entry in the document’s trailer
            # dictionary; see Table 15) to the hash function and finish the hash.
            h.update(self._document_id)
            digest: bytes = h.digest()

            # d) Encrypt the 16-byte result of the hash, using an RC4 encryption function with the encryption key from step
            # (a).
            digest = RC4().encrypt(key_rev_3, digest)

            # e) Do the following 19 times: Take the output from the previous invocation of the RC4 function and pass it as
            # input to a new invocation of the function; use an encryption key generated by taking each byte of the
            # original encryption key obtained in step (a) and performing an XOR (exclusive or) operation between that
            # byte and the single-byte value of the iteration counter (from 1 to 19).
            if self._revision >= 3:
                for i in range(1, 20):
                    key2: bytes = bytes([b ^ i for b in key_rev_3])
                    digest = RC4().encrypt(key2, digest)

            # f) Append 16 bytes of arbitrary padding to the output from the final invocation of the RC4 function and store
            # the 32-byte result as the value of the U entry in the encryption dictionary.
            digest += bytes(
                [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53]
            )
            self._u = digest

    def _compute_encryption_key(self, password: typing.Optional[bytes]) -> bytes:
        # a) Pad or truncate the password string to exactly 32 bytes. If the password string is more than 32 bytes long,
        # use only its first 32 bytes; if it is less than 32 bytes long, pad it by appending the required number of
        # additional bytes from the beginning of the following padding string:
        # < 28 BF 4E 5E 4E 75 8A 41 64 00 4E 56 FF FA 01 08
        # 2E 2E 00 B6 D0 68 3E 80 2F 0C A9 FE 64 53 69 7A >
        # That is, if the password string is n bytes long, append the first 32 - n bytes of the padding string to the end
        # of the password string. If the password string is empty (zero-length), meaning there is no user password,
        # substitute the entire padding string in its place.

        # b) Initialize the MD5 hash function and pass the result of step (a) as input to this function.
        h = hashlib.md5()
        h.update(StandardSecurityHandler._pad_or_truncate(password))

        # c) Pass the value of the encryption dictionary’s O entry to the MD5 hash function. ("Algorithm 3: Computing
        # the encryption dictionary’s O (owner password) value" shows how the O value is computed.)
        h.update(self._o)

        # d) Convert the integer value of the P entry to a 32-bit unsigned binary number and pass these bytes to the
        # MD5 hash function, low-order byte first.
        h.update(self._permissions.to_bytes(length=4, byteorder="little", signed=True))

        # e) Pass the first element of the file’s file identifier array (the value of the ID entry in the document’s trailer
        # dictionary; see Table 15) to the MD5 hash function.
        h.update(self._document_id)

        # f) (Security handlers of revision 4 or greater) If document metadata is not being encrypted, pass 4 bytes with
        # the value 0xFFFFFFFF to the MD5 hash function.
        if self._revision >= 4 and not self._encrypt_metadata:
            h.update(bytes([255, 255, 255, 255]))

        # g) Finish the hash.
        digest: bytes = h.digest()

        # h) (Security handlers of revision 3 or greater) Do the following 50 times: Take the output from the previous
        # MD5 hash and pass the first n bytes of the output as input into a new MD5 hash, where n is the number of
        # bytes of the encryption key as defined by the value of the encryption dictionary’s Length entry.
        n: int = 0
        if self._revision >= 3:
            n = int(self._key_length / 8)
            for _ in range(0, 50):
                h2 = hashlib.md5()
                h2.update(digest[0:n])
                digest = h2.digest()

        # i) Set the encryption key to the first n bytes of the output from the final MD5 hash, where n shall always be 5
        # for security handlers of revision 2 but, for security handlers of revision 3 or greater, shall depend on the
        # value of the encryption dictionary’s Length entry.
        n = 5
        if self._revision >= 3:
            n = int(self._key_length / 8)
        encryption_key: bytes = digest[0:n]

        return encryption_key

    def _decrypt_data(self, object: AnyPDFType) -> AnyPDFType:
        return self._encrypt_data(object)

    def _encrypt_data(self, object: AnyPDFType) -> AnyPDFType:
        # a) Obtain the object number and generation number from the object identifier of the string or stream to be
        # encrypted (see 7.3.10, "Indirect Objects"). If the string is a direct object, use the identifier of the indirect
        # object containing it.
        reference: typing.Optional[Reference] = object.get_reference()
        if reference is None:
            parent: typing.Optional[PDFObject] = object.get_parent()
            assert parent is not None
            reference = parent.get_reference()
        assert reference is not None
        assert reference.object_number is not None
        assert reference.generation_number is not None
        object_number: int = reference.object_number
        generation_number: int = reference.generation_number

        # b) For all strings and streams without crypt filter specifier; treating the object number and generation number
        # as binary integers, extend the original n-byte encryption key to n + 5 bytes by appending the low-order 3
        # bytes of the object number and the low-order 2 bytes of the generation number in that order, low-order byte
        # first. (n is 5 unless the value of V in the encryption dictionary is greater than 1, in which case n is the value
        # of Length divided by 8.)
        # If using the AES algorithm, extend the encryption key an additional 4 bytes by adding the value “sAlT”,
        # which corresponds to the hexadecimal values 0x73, 0x41, 0x6C, 0x54. (This addition is done for backward
        # compatibility and is not intended to provide additional security.)
        encryption_key = (
            self._encryption_key
            + object_number.to_bytes(3, byteorder="little", signed=False)
            + generation_number.to_bytes(2, byteorder="little", signed=False)
        )
        n: int = 5
        if self._v > 1:
            n = int(self._key_length / 8)

        # c) Initialize the MD5 hash function and pass the result of step (b) as input to this function.
        h = hashlib.md5()
        h.update(encryption_key)

        # d) Use the first (n + 5) bytes, up to a maximum of 16, of the output from the MD5 hash as the key for the RC4
        # or AES symmetric key algorithms, along with the string or stream data to be encrypted.
        # If using the AES algorithm, the Cipher Block Chaining (CBC) mode, which requires an initialization vector,
        # is used. The block size parameter is set to 16 bytes, and the initialization vector is a 16-byte random
        # number that is stored as the first 16 bytes of the encrypted stream or string.
        # The output is the encrypted data to be stored in the PDF file.
        n_plus_5: int = min(16, n + 5)
        if isinstance(object, String):
            str_new_content_bytes: bytes = RC4().encrypt(
                h.digest()[0:n_plus_5], object.get_content_bytes()
            )
            # TODO
        if isinstance(object, HexadecimalString):
            hex_str_new_content_bytes: bytes = RC4().encrypt(
                h.digest()[0:n_plus_5], object.get_content_bytes()
            )
            # TODO
        if isinstance(object, Stream):
            stream_new_content_bytes: bytes = RC4().encrypt(
                h.digest()[0:n_plus_5], object["DecodedBytes"]
            )
            object[Name("DecodedBytes")] = stream_new_content_bytes
            object[Name("Bytes")] = zlib.compress(object["DecodedBytes"], 9)
            return object

        # default
        return object

    @staticmethod
    def _pad_or_truncate(b: typing.Optional[bytes]) -> bytes:
        # fmt: off
        padding: bytes = bytes([40, 191, 78, 94, 78, 117, 138, 65,
                                100, 0, 78, 86, 255, 250, 1, 8,
                                46, 46, 0, 182, 208, 104, 62, 128,
                                47, 12, 169, 254, 100, 83, 105, 122])
        # fmt: on
        if b is None:
            return padding
        if len(b) > 32:
            return b[0:32]
        if len(b) < 32:
            b2: bytes = b + padding
            return b2[0:32]
        return b

    @staticmethod
    def _str_to_bytes(s: typing.Optional[str]) -> typing.Optional[bytes]:
        if s is None:
            return None
        return bytes(s, encoding="charmap")

    @staticmethod
    def _unescape_pdf_syntax(
        s: typing.Union[str, String, None]
    ) -> typing.Optional[str]:
        # None
        if s is None:
            return None
        # String
        if isinstance(s, String):
            return str(s.get_content_bytes(), encoding="latin1")
        # str
        return str(String(s).get_content_bytes(), encoding="latin1")

    #
    # PUBLIC
    #



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A screenshot (also known as screen capture or screen grab) is a digital image that shows the contents of a computer display.
A screenshot is created by the operating system or software running on the device powering the display.
"""
import typing
from _decimal import Decimal

from PIL import ImageGrab

from borb.pdf import Image, Alignment, Color, HexColor
from borb.pdf.canvas.geometry.rectangle import Rectangle


class ScreenShot(Image):
    """
    A screenshot (also known as screen capture or screen grab) is a digital image that shows the contents of a computer display.
    A screenshot is created by the operating system or software running on the device powering the display.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        all_screens: bool = False,
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        bounding_box: typing.Optional[Rectangle] = None,
        height: typing.Optional[Decimal] = None,
        horizontal_alignment: Alignment = Alignment.LEFT,
        include_layered_windows: bool = False,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_top: typing.Optional[Decimal] = None,
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        vertical_alignment: Alignment = Alignment.TOP,
        width: typing.Optional[Decimal] = None,
        x_display: typing.Optional[str] = None,
    ):
        super().__init__(
            image=ImageGrab.grab(
                bbox=None
                if bounding_box is None
                else (
                    bounding_box.get_x(),
                    bounding_box.get_y(),
                    bounding_box.get_width(),
                    bounding_box.get_height(),
                ),
                include_layered_windows=include_layered_windows,
                all_screens=all_screens,
                xdisplay=x_display,
            ),
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            height=height,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
            width=width,
        )

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener  extracts the colors used in rendering a PDF
"""
import io
import typing
from decimal import Decimal

from PIL.Image import Image  # type: ignore [import]

from borb.pdf.document.document import Document
from borb.pdf.canvas.canvas import Canvas
from borb.pdf.canvas.canvas_stream_processor import CanvasStreamProcessor
from borb.pdf.canvas.color.color import RGBColor, Color, HSVColor
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.chunk_of_text_render_event import ChunkOfTextRenderEvent
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.page.page import Page


class ColorExtraction(EventListener):
    """
    This implementation of EventListener  extracts the colors used in rendering a PDF
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        max_number_of_colors_to_return: int = 32,
        max_number_of_colors_to_register: int = 32,
    ):
        """
        Constructs a new ColorSpectrumExtraction
        """
        assert (
            max_number_of_colors_to_return >= 0
        ), "max_number_of_colors_to_return must be a positive integer"
        assert (
            max_number_of_colors_to_register >= 0
        ), "max_number_of_colors_to_register must be a positive integer"
        self._max_number_of_colors_to_return: int = max_number_of_colors_to_return
        self._max_number_of_colors_to_register: int = max_number_of_colors_to_register

        # build color_palette
        self._colors_from_palette_per_page: typing.Dict[
            int, typing.Dict[Color, Decimal]
        ] = {}
        self._color_palette: typing.List[Color] = [
            HSVColor(Decimal(x / 360), Decimal(1), Decimal(1)).to_rgb()
            for x in range(0, 360, self._max_number_of_colors_to_register)
        ]

        # keep track of current_page
        self._current_page: int = -1

    #
    # PRIVATE
    #

    def _begin_page(self, page: Page):
        self._current_page += 1
        self._colors_from_palette_per_page[self._current_page] = {}

    @staticmethod
    def _color_distance(c0: Color, c1: Color) -> Decimal:
        rgb0: RGBColor = c0.to_rgb()
        rgb1: RGBColor = c1.to_rgb()
        return (
            (rgb0.red - rgb1.red) ** 2
            + (rgb0.green - rgb1.green) ** 2
            + (rgb0.blue - rgb1.blue) ** 2
        )

    def _end_page(self, page: Page):
        vs: typing.List[typing.Tuple[Color, Decimal]] = [
            (k, v)
            for k, v in self._colors_from_palette_per_page[self._current_page].items()
        ]
        vs.sort(key=lambda x: x[1], reverse=True)
        self._colors_from_palette_per_page[self._current_page] = {
            k: v for k, v in vs[0 : self._max_number_of_colors_to_return]
        }

    def _event_occurred(self, event: Event) -> None:
        if isinstance(event, BeginPageEvent):
            self._begin_page(event.get_page())
        if isinstance(event, EndPageEvent):
            self._end_page(event.get_page())
        if isinstance(event, ChunkOfTextRenderEvent):
            self._render_text(event)
        if isinstance(event, ImageRenderEvent):
            self._render_image(event)

    @staticmethod
    def _get_rgb_from_image(img: Image, x: int, y: int) -> RGBColor:
        c = img.getpixel((x, y))
        if img.mode == "RGB":
            return RGBColor(
                r=Decimal(c[0] / 255),
                g=Decimal(c[1] / 255),
                b=Decimal(c[2] / 255),
            )
        if img.mode == "RGBA":
            return RGBColor(
                r=Decimal(c[0] / 255),
                g=Decimal(c[1] / 255),
                b=Decimal(c[2] / 255),
            )
        if img.mode == "CMYK":
            r = int((1 - c[0]) * (1 - c[3]))
            g = int((1 - c[1]) * (1 - c[3]))
            b = int((1 - c[2]) * (1 - c[0]))
            return RGBColor(r=Decimal(r), g=Decimal(g), b=Decimal(b))
        return RGBColor(Decimal(0), Decimal(0), Decimal(0))

    @staticmethod
    def _nearest_color(c: Color, color_palette: typing.List[Color]) -> Color:
        c1: Color = color_palette[0]
        for c2 in color_palette:
            d1 = ColorExtraction._color_distance(c, c1)
            d2 = ColorExtraction._color_distance(c, c2)
            if d2 < d1:
                c1 = c2
        return c1

    def _render_image(self, event: ImageRenderEvent):
        r = (event.get_width() * event.get_height()) / (event.get_image().width * event.get_image().height)  # type: ignore[attr-defined]
        color_count: typing.Dict[RGBColor, Decimal] = {}
        for i in range(0, event.get_image().width):  # type: ignore [attr-defined]
            for j in range(0, event.get_image().height):  # type: ignore [attr-defined]
                c = ColorExtraction._get_rgb_from_image(event.get_image(), i, j)
                if c not in color_count:
                    color_count[c] = Decimal(1)
                else:
                    color_count[c] += Decimal(1)
        for k, v in color_count.items():
            self._update_color_count(v * r, k)

    def _render_text(self, event: ChunkOfTextRenderEvent):
        assert event is not None
        bb: typing.Optional[Rectangle] = event.get_previous_layout_box()
        s: Decimal = Decimal(0) if bb is None else (bb.width * bb.height)
        c: RGBColor = event.get_font_color().to_rgb()
        self._update_color_count(s, c)

    def _update_color_count(self, amount: Decimal, color: RGBColor):
        c2: Color = ColorExtraction._nearest_color(color, self._color_palette)
        self._colors_from_palette_per_page[self._current_page][c2] = (
            self._colors_from_palette_per_page[self._current_page].get(c2, Decimal(0))
            + amount
        )

    #
    # PUBLIC
    #

    def get_color(
        self,
    ) -> typing.Dict[int, typing.Dict[Color, Decimal]]:
        """
        This function returns the colors used on a given page of a PDF
        """
        return self._colors_from_palette_per_page

    @staticmethod
    def get_color_from_pdf(
        pdf: Document,
        max_number_of_colors_to_return: int = 32,
        max_number_of_colors_to_register: int = 32,
    ) -> typing.Dict[int, typing.Dict[Color, Decimal]]:
        """
        This function returns the colors used in a given PDF
        :param pdf:                                 the PDF to be analysed
        :param max_number_of_colors_to_return:      the maximum number of colors to return (i.e. "the top 10")
        :param max_number_of_colors_to_register:    the maximum number of colors to register (i.e. "do not distinguish between green and dark-green")
        :return:                                    the colors used in a given PDF
        """
        colors_per_page: typing.Dict[int, typing.Dict[Color, Decimal]] = {}
        number_of_pages: int = int(pdf.get_document_info().get_number_of_pages() or 0)
        for page_nr in range(0, number_of_pages):
            # get Page object
            page: Page = pdf.get_page(page_nr)
            page_source: io.BytesIO = io.BytesIO(page["Contents"]["DecodedBytes"])
            # register EventListener
            cse: "ColorExtraction" = ColorExtraction(
                max_number_of_colors_to_return=max_number_of_colors_to_return,
                max_number_of_colors_to_register=max_number_of_colors_to_register,
            )
            # process Page
            cse._event_occurred(BeginPageEvent(page))
            CanvasStreamProcessor(page, Canvas(), []).read(page_source, [cse])
            cse._event_occurred(EndPageEvent(page))
            # add to output dictionary
            colors_per_page[page_nr] = cse.get_color()[0]
        # return
        return colors_per_page



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    In information retrieval, tf–idf, TF*IDF, or TFIDF, short for term frequency–inverse document frequency,
    is a numerical statistic that is intended to reflect how important a word is to a document in a collection or corpus.
    It is often used as a weighting factor in searches of information retrieval, text mining, and user modeling.

    The tf–idf value increases proportionally to the number of times a word appears in the document
    and is offset by the number of documents in the corpus that contain the word,
    which helps to adjust for the fact that some words appear more frequently in general.

    tf–idf is one of the most popular term-weighting schemes today.
    A survey conducted in 2015 showed that 83% of text-based recommender systems in digital libraries use tf–idf.

    Variations of the tf–idf weighting scheme are often used by search engines as a central tool
    in scoring and ranking a document's relevance given a user query.
    tf–idf can be successfully used for stop-words filtering in various subject fields,
    including text summarization and classification.

    One of the simplest ranking functions is computed by summing the tf–idf for each query term;
    many more sophisticated ranking functions are variants of this simple model.
"""
import io
import re
import typing
from math import log

from borb.pdf.document.document import Document
from borb.pdf.canvas.canvas import Canvas
from borb.pdf.canvas.canvas_stream_processor import CanvasStreamProcessor
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.page.page import Page
from borb.toolkit.text.simple_text_extraction import SimpleTextExtraction


class TFIDFKeywordExtraction(SimpleTextExtraction):
    """
    In information retrieval, tf–idf, TF*IDF, or TFIDF, short for term frequency–inverse document frequency,
    is a numerical statistic that is intended to reflect how important a word is to a document in a collection or corpus.
    It is often used as a weighting factor in searches of information retrieval, text mining, and user modeling.

    The tf–idf value increases proportionally to the number of times a word appears in the document
    and is offset by the number of documents in the corpus that contain the word,
    which helps to adjust for the fact that some words appear more frequently in general.

    tf–idf is one of the most popular term-weighting schemes today.
    A survey conducted in 2015 showed that 83% of text-based recommender systems in digital libraries use tf–idf.

    Variations of the tf–idf weighting scheme are often used by search engines as a central tool
    in scoring and ranking a document's relevance given a user query.
    tf–idf can be successfully used for stop-words filtering in various subject fields,
    including text summarization and classification.

    One of the simplest ranking functions is computed by summing the tf–idf for each query term;
    many more sophisticated ranking functions are variants of this simple model.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self, stopwords: typing.List[str] = [], minimum_term_frequency: int = 3
    ):
        super().__init__()
        self._term_frequency_per_page: typing.Dict[int, typing.Dict[str, float]] = {}
        self._number_of_words_per_page: typing.Dict[int, float] = {}
        self._inverse_page_frequency: typing.Dict[str, float] = {}
        self._stopwords: typing.List[str] = [x.upper() for x in stopwords]
        self._number_of_pages: float = 0
        self._minimum_term_frequency: float = minimum_term_frequency

    #
    # PRIVATE
    #

    def _end_page(self, page: Page):
        super()._end_page(page)

        # update number of pages
        self._number_of_pages += 1

        # get words
        words_on_page: typing.List[str] = [
            x.upper()
            for x in re.split("[^a-zA-Z]+", self.get_text()[self._current_page])
            if x.upper() not in self._stopwords
        ]

        # update _term_frequency_per_page
        self._term_frequency_per_page[self._current_page] = {}
        self._number_of_words_per_page[self._current_page] = len(words_on_page)
        for w in words_on_page:
            self._term_frequency_per_page[self._current_page][w] = (
                self._term_frequency_per_page[self._current_page].get(w, 0) + 1
            )

        # update _inverse_page_frequency
        for w in set(words_on_page):
            self._inverse_page_frequency[w] = self._inverse_page_frequency.get(w, 0) + 1

    #
    # PUBLIC
    #

    def get_keywords(self) -> typing.Dict[int, typing.List[typing.Tuple[str, float]]]:
        """
        This function returns a typing.List[typing.Tuple[str, float]] for a given page
        """
        out: typing.Dict[int, typing.List[typing.Tuple[str, float]]] = {}
        for k, v in self._term_frequency_per_page.items():
            if k not in out:
                out[k] = []
            for w, tf in v.items():

                # check minimum_term_frequency
                if tf < self._minimum_term_frequency:
                    continue

                # normalize tf, idf
                tf /= self._number_of_words_per_page[k]
                idf = log(self._number_of_pages / self._inverse_page_frequency[w])

                # avoid multiply by zero
                tf += 0.0001
                idf += 0.0001

                # calculate tf-idf score
                out[k].append((w, tf * idf))

            # sort
            out[k].sort(key=lambda x: x[1], reverse=True)

        # return
        return out

    @staticmethod
    def get_keywords_from_pdf(
        pdf: Document,
    ) -> typing.Dict[int, typing.List[typing.Tuple[str, float]]]:
        """
        This function returns the keywords for a given PDF (per page)
        :param pdf:     the PDF to be analyzed
        :return:        the keywords per page (represented by typing.Dict[int, typing.List[typing.Tuple[str, float]]])
        """
        keywords_per_page: typing.Dict[int, typing.List[typing.Tuple[str, float]]] = {}
        number_of_pages: int = int(pdf.get_document_info().get_number_of_pages() or 0)
        for page_nr in range(0, number_of_pages):
            # get Page object
            page: Page = pdf.get_page(page_nr)
            page_source: io.BytesIO = io.BytesIO(page["Contents"]["DecodedBytes"])
            # register EventListener
            l: "TFIDFKeywordExtraction" = TFIDFKeywordExtraction()
            # process Page
            l._event_occurred(BeginPageEvent(page))
            CanvasStreamProcessor(page, Canvas(), []).read(page_source, [l])
            l._event_occurred(EndPageEvent(page))
            # add to output dictionary
            keywords_per_page[page_nr] = l.get_keywords()[0]
        # return
        return keywords_per_page



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of FormField represents a push button.
"""
import typing
import zlib
from decimal import Decimal

from borb.io.read.pdf_object import PDFObject
from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, Name, List, String, Stream, Boolean
from borb.pdf.canvas.color.color import HexColor, Color
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.forms.form_field import FormField
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.canvas.layout.text.line_of_text import LineOfText
from borb.pdf.page.page import Page


class PushButton(FormField):
    """
    This implementation of FormField represents a push button.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        text: str,
        background_color: typing.Optional[Color] = HexColor("efefef"),
        border_bottom: bool = True,
        border_color: Color = HexColor("767676"),
        border_left: bool = True,
        border_right: bool = True,
        border_top: bool = True,
        border_width: Decimal = Decimal(1),
        field_name: typing.Optional[str] = None,
        font_size: typing.Optional[Decimal] = Decimal(12),
        font_color: Color = HexColor("000000"),
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: typing.Optional[Decimal] = Decimal(0),
        margin_left: typing.Optional[Decimal] = Decimal(0),
        margin_right: typing.Optional[Decimal] = Decimal(0),
        margin_top: typing.Optional[Decimal] = Decimal(0),
        padding_bottom: Decimal = Decimal(2),
        padding_left: Decimal = Decimal(6),
        padding_right: Decimal = Decimal(6),
        padding_top: Decimal = Decimal(2),
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        super(PushButton, self).__init__(
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=Decimal(0),
            border_radius_bottom_right=Decimal(0),
            border_radius_top_left=Decimal(0),
            border_radius_top_right=Decimal(0),
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            font_size=font_size,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )
        assert len(text) > 0
        self._field_name: typing.Optional[str] = field_name
        self._text = text
        self._font_color: Color = font_color
        self._widget_dictionary: typing.Optional[Dictionary] = None

    #
    # PRIVATE
    #

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        assert self._font_size is not None
        line_of_text: LineOfText = LineOfText(
            self._text,
            background_color=self._background_color,
            font_size=self._font_size,
            font_color=self._font_color,
        )
        return line_of_text.get_layout_box(available_space)

    def _init_widget_dictionary(self, page: Page) -> None:
        if self._widget_dictionary is not None:
            return

        root: typing.Optional[PDFObject] = page.get_root()
        assert root is not None
        assert isinstance(root, Dictionary)
        if "XRef" not in root:
            return

        # init page and font resources
        assert self._font_size is not None
        font_resource_name: Name = self._get_font_resource_name(
            StandardType1Font("Helvetica"), page
        )

        # widget resource dictionary
        widget_resources: Dictionary = Dictionary()
        widget_resources.set_is_unique(True)
        widget_resources[Name("Font")] = page["Resources"]["Font"]

        # get Catalog
        catalog: Dictionary = root["XRef"]["Trailer"]["Root"]

        # widget dictionary
        # fmt: off
        self._widget_dictionary = Dictionary()
        self._widget_dictionary.set_is_unique(True)
        self._widget_dictionary[Name("AA")] = Dictionary()
        self._widget_dictionary[Name("AA")][Name("D")] = Dictionary()
        self._widget_dictionary[Name("AA")][Name("D")][Name("Type")] = Name("Action")
        self._widget_dictionary[Name("AA")][Name("D")][Name("S")] = Name("ResetForm")
        # fmt: on

        # create normal appearance
        # fmt: off
        self._widget_dictionary[Name("AP")] = Dictionary()
        self._widget_dictionary[Name("AP")].set_is_unique(True)
        self._widget_dictionary[Name("AP")][Name("N")] = Stream()
        self._widget_dictionary[Name("AP")][Name("N")][Name("Type")] = Name("XObject")
        self._widget_dictionary[Name("AP")][Name("N")][Name("Subtype")] = Name("Form")
        self._widget_dictionary[Name("AP")][Name("N")][Name("BBox")] = List().set_is_inline(True)   # type: ignore[attr-defined]
        for _ in range(0, 4):
            self._widget_dictionary[Name("AP")][Name("N")][Name("BBox")].append(bDecimal(0))
        self._widget_dictionary[Name("AP")][Name("N")][Name("DecodedBytes")] = b"/Tx BMC EMC"
        self._widget_dictionary[Name("AP")][Name("N")][Name("Bytes")] = zlib.compress(self._widget_dictionary[Name("AP")][Name("N")][Name("DecodedBytes")], 9)
        self._widget_dictionary[Name("AP")][Name("N")][Name("Filter")] = Name("FlateDecode")
        self._widget_dictionary[Name("AP")][Name("N")][Name("Length")] = bDecimal(len(self._widget_dictionary[Name("AP")][Name("N")][Name("Bytes")]))
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")] = Dictionary()
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("ProcSet")] = List().set_is_inline(True) # type: ignore [attr-defined]
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("ProcSet")].append(Name("PDF"))
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("ProcSet")].append(Name("Text"))
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("Font")] = Dictionary()
        self._widget_dictionary[Name("AP")][Name("N")][Name("Resources")][Name("Font")] = page["Resources"]["Font"]
        # fmt: on

        # create default appearance
        # fmt: off
        self._widget_dictionary[Name("DA")] = String("0.23921 0.23921 0.23921 rg /%s %f Tf" % (font_resource_name, float(self._font_size)))
        self._widget_dictionary[Name("DR")] = widget_resources
        # fmt: on

        # other flags
        # fmt: off
        self._widget_dictionary[Name("F")] = bDecimal(4)
        self._widget_dictionary[Name("Ff")] = bDecimal(65536)
        self._widget_dictionary[Name("FT")] = Name("Btn")
        self._widget_dictionary[Name("MK")] = Dictionary()
        self._widget_dictionary[Name("MK")][Name("BC")] = List().set_is_inline(True)    # type: ignore [attr-defined]
        self._widget_dictionary[Name("MK")][Name("BG")] = List().set_is_inline(True)    # type: ignore [attr-defined]
        self._widget_dictionary[Name("MK")][Name("CA")] = String("")
        self._widget_dictionary[Name("P")] = catalog
        self._widget_dictionary[Name("Q")] = bDecimal(1)
        self._widget_dictionary[Name("Rect")] = List().set_is_inline(True)              # type: ignore [attr-defined]
        for _ in range(0, 4):
            self._widget_dictionary[Name("Rect")].append(bDecimal(0))
        self._widget_dictionary[Name("Subtype")] = Name("Widget")
        self._widget_dictionary[Name("T")] = String(self._field_name or self._get_auto_generated_field_name(page))
        self._widget_dictionary[Name("Type")] = Name("Annot")
        # fmt: on

        # append field to page /Annots
        if "Annots" not in page:
            page[Name("Annots")] = List()
        page["Annots"].append(self._widget_dictionary)

        # append field to catalog
        if "AcroForm" not in catalog:
            catalog[Name("AcroForm")] = Dictionary()
            catalog["AcroForm"][Name("Fields")] = List()
            catalog["AcroForm"][Name("DR")] = widget_resources
            catalog["AcroForm"][Name("NeedAppearances")] = Boolean(True)
        catalog["AcroForm"]["Fields"].append(self._widget_dictionary)

    def _paint_content_box(self, page: "Page", available_space: Rectangle) -> None:

        # init self._widget_dictionary
        self._init_widget_dictionary(page)

        # layout text
        assert self._font_size is not None
        line_of_text = LineOfText(
            self._text,
            background_color=self._background_color,
            font_size=self._font_size,
            font_color=self._font_color,
        )
        line_of_text.paint(page, self._get_content_box(available_space))
        cbox: typing.Optional[Rectangle] = line_of_text.get_previous_paint_box()
        assert cbox is not None

        # add our own margins to expand the clickable part
        cbox.x -= self._padding_left
        cbox.width += self._padding_left + self._padding_right
        cbox.y -= self._padding_bottom
        cbox.height += self._padding_bottom + self._padding_top

        # set location
        # fmt: off
        if self._widget_dictionary is not None:
            self._widget_dictionary["Rect"][0] = bDecimal(cbox.get_x())                             # ll_x
            self._widget_dictionary["Rect"][1] = bDecimal(cbox.get_y())                             # ll_y
            self._widget_dictionary["Rect"][2] = bDecimal(cbox.get_x() + cbox.get_width())          # ur_x
            self._widget_dictionary["Rect"][3] = bDecimal(cbox.get_y() + cbox.get_height())         # ur_y
        # fmt: on

    #
    # PUBLIC
    #


class JavaScriptPushButton(PushButton):
    """
    This implementation of FormField represents a push button that triggers JavaScript.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        javascript: str,
        text: str,
        background_color: typing.Optional[Color] = HexColor("efefef"),
        border_bottom: bool = True,
        border_color: Color = HexColor("767676"),
        border_left: bool = True,
        border_right: bool = True,
        border_top: bool = True,
        border_width: Decimal = Decimal(1),
        field_name: typing.Optional[str] = None,
        font_size: typing.Optional[Decimal] = Decimal(12),
        font_color: Color = HexColor("000000"),
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: Decimal = Decimal(0),
        margin_left: Decimal = Decimal(0),
        margin_right: Decimal = Decimal(0),
        margin_top: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(2),
        padding_left: Decimal = Decimal(6),
        padding_right: Decimal = Decimal(6),
        padding_top: Decimal = Decimal(2),
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        super(JavaScriptPushButton, self).__init__(
            text=text,
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            field_name=field_name,
            font_size=font_size,
            font_color=font_color,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )
        self._javascript: str = javascript

    #
    # PRIVATE
    #

    def _init_widget_dictionary(self, page: Page) -> None:

        # call to super
        super(JavaScriptPushButton, self)._init_widget_dictionary(page)

        # build JavaScript stream object
        # fmt: off
        javascript_stream = Stream()
        javascript_stream.set_is_unique(True)     # type: ignore [attr-defined]
        javascript_stream[Name("Type")] = Name("JavaScript")
        javascript_stream[Name("DecodedBytes")] = bytes(self._javascript, "latin1")
        javascript_stream[Name("Bytes")] = zlib.compress(javascript_stream[Name("DecodedBytes")], 9)
        javascript_stream[Name("Length")] = bDecimal(len(javascript_stream[Name("Bytes")]))
        javascript_stream[Name("Filter")] = Name("FlateDecode")
        # fmt: on

        # modify action dictionary of PushButton (super)
        # fmt: off
        if self._widget_dictionary is not None:
            self._widget_dictionary[Name("AA")][Name("D")][Name("S")] = Name("JavaScript")
            self._widget_dictionary[Name("AA")][Name("D")][Name("JS")] = javascript_stream
        # fmt: on

    #
    # PUBLIC
    #



# !/usr/bin/env python
# -*- coding: utf-8 -*-
"""
In computer science, lexical analysis, lexing or tokenization is the process of converting a sequence of characters
(such as in a computer program or web page) into a sequence of tokens (strings with an assigned and thus identified meaning).
A program that performs lexical analysis may be termed a lexer, tokenizer, or scanner,
although scanner is also a term for the first stage of a lexer.
A lexer is generally combined with a parser, which together analyze the syntax of programming languages, web pages,
and so forth.
"""
import typing

from borb.pdf.canvas.layout.equation.token import Token, TokenType


class Tokenizer:
    """
    In computer science, lexical analysis, lexing or tokenization is the process of converting a sequence of characters
    (such as in a computer program or web page) into a sequence of tokens (strings with an assigned and thus identified meaning).
    A program that performs lexical analysis may be termed a lexer, tokenizer, or scanner,
    although scanner is also a term for the first stage of a lexer.
    A lexer is generally combined with a parser, which together analyze the syntax of programming languages, web pages,
    and so forth.
    """

    @staticmethod
    def tokenize(infix_expression: str) -> typing.List[Token]:
        """
        This function converts an infix expression to a typing.List[Token]
        :param infix_expression:    the infix expression (as str) to tokenize
        :return:                    the typing.List[Token] representing the tokenized output
        """
        # variable to store output
        output: typing.List[Token] = []

        # process expression
        i: int = 0
        while i < len(infix_expression):

            #
            # SPACE
            #
            if infix_expression[i] == " ":
                i += 1
                continue

            #
            # PARENTHESES
            #

            if infix_expression[i] == "(":
                output += [Token(text="(", type=TokenType.LEFT_PARENTHESIS)]
                i += 1
                continue
            if infix_expression[i] == ")":
                output += [Token(text=")", type=TokenType.RIGHT_PARENTHESIS)]
                i += 1
                continue

            #
            # BASIC OPERATORS
            #

            if infix_expression[i] == "±":
                output += [
                    Token(
                        text="±",
                        type=TokenType.OPERATOR,
                        precedence=2,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue
            if infix_expression[i] == "+":
                output += [
                    Token(
                        text="+",
                        type=TokenType.OPERATOR,
                        precedence=2,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue
            if infix_expression[i] == "-":
                output += [
                    Token(
                        text="-",
                        type=TokenType.OPERATOR,
                        precedence=2,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue
            if infix_expression[i] == "*":
                output += [
                    Token(
                        text="*",
                        type=TokenType.OPERATOR,
                        precedence=3,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue
            if infix_expression[i] == "/":
                output += [
                    Token(
                        text="/",
                        type=TokenType.OPERATOR,
                        precedence=3,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue

            #
            # COMPARISON
            #

            if infix_expression[i:].startswith("!="):
                output += [
                    Token(
                        text="≠",
                        type=TokenType.OPERATOR,
                        precedence=1,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 2
                continue

            if infix_expression[i:].startswith(">="):
                output += [
                    Token(
                        text="≥",
                        type=TokenType.OPERATOR,
                        precedence=1,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 2
                continue

            if infix_expression[i:].startswith("<="):
                output += [
                    Token(
                        text="≤",
                        type=TokenType.OPERATOR,
                        precedence=1,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 2
                continue

            if infix_expression[i] == ">":
                output += [
                    Token(
                        text=">",
                        type=TokenType.OPERATOR,
                        precedence=1,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue

            if infix_expression[i] == "<":
                output += [
                    Token(
                        text="<",
                        type=TokenType.OPERATOR,
                        precedence=1,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue

            if infix_expression[i] == "=":
                output += [
                    Token(
                        text="=",
                        type=TokenType.OPERATOR,
                        precedence=1,
                        is_left_associative=True,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue

            #
            # POWERS
            #

            if infix_expression[i] == "^":
                output += [
                    Token(
                        text="^",
                        type=TokenType.OPERATOR,
                        precedence=4,
                        is_left_associative=False,
                        number_of_arguments=2,
                    )
                ]
                i += 1
                continue
            if infix_expression[i] == "²":
                output += [
                    Token(
                        text="^",
                        type=TokenType.OPERATOR,
                        precedence=4,
                        is_left_associative=False,
                        number_of_arguments=2,
                    )
                ]
                output += [Token(text="2", type=TokenType.NUMBER)]
                i += 1
                continue
            if infix_expression[i] == "³":
                output += [
                    Token(
                        text="^",
                        type=TokenType.OPERATOR,
                        precedence=4,
                        is_left_associative=False,
                        number_of_arguments=2,
                    )
                ]
                output += [Token(text="3", type=TokenType.NUMBER)]
                i += 1
                continue
            if infix_expression[i:].startswith("sqrt"):
                output += [
                    Token(
                        text="sqrt",
                        type=TokenType.FUNCTION,
                        precedence=4,
                        is_left_associative=False,
                        number_of_arguments=1,
                    )
                ]
                i += 4
                continue
            #
            # TRIG FUNCTIONS
            #

            if infix_expression[i:].startswith("sin"):
                output += [
                    Token(text="sin", type=TokenType.FUNCTION, number_of_arguments=1)
                ]
                i += 3
                continue
            if infix_expression[i:].startswith("cos"):
                output += [
                    Token(text="cos", type=TokenType.FUNCTION, number_of_arguments=1)
                ]
                i += 3
                continue
            if infix_expression[i:].startswith("tan"):
                output += [
                    Token(text="tan", type=TokenType.FUNCTION, number_of_arguments=1)
                ]
                i += 3
                continue
            if infix_expression[i:].startswith("cot"):
                output += [
                    Token(text="cot", type=TokenType.FUNCTION, number_of_arguments=1)
                ]
                i += 3
                continue
            if infix_expression[i:].startswith("sec"):
                output += [
                    Token(text="sec", type=TokenType.FUNCTION, number_of_arguments=1)
                ]
                i += 3
                continue
            if infix_expression[i:].startswith("csc"):
                output += [
                    Token(text="csc", type=TokenType.FUNCTION, number_of_arguments=1)
                ]
                i += 3
                continue

            #
            # ABS
            #

            if infix_expression[i:].startswith("abs"):
                output += [
                    Token(text="abs", type=TokenType.FUNCTION, number_of_arguments=1)
                ]
                i += 3
                continue

            #
            # NUMBERS
            #

            if infix_expression[i] in "0123456789.":
                j: int = i
                while (
                    j < len(infix_expression) and infix_expression[j] in "0123456789."
                ):
                    j += 1
                output += [Token(text=infix_expression[i:j], type=TokenType.NUMBER)]
                i = j
                continue

            # IF the token can not be interpreted as anything else
            # AND the previous token was a VARIABLE
            # THEN concatenate the token to the previous VARIABLE
            if len(output) > 0 and output[-1]._type == TokenType.VARIABLE:
                output[-1]._text += infix_expression[i]
                i += 1
                continue

            # IF the token can not be interpreted as anything else
            # AND there is no previous token
            # THEN mark the token as a variable
            output += [Token(text=infix_expression[i], type=TokenType.VARIABLE)]
            i += 1

        # fix unary minus
        for i, t in enumerate(output):
            if output[i].get_type() != TokenType.OPERATOR:
                continue
            if output[i].get_text() != "-":
                continue
            if (
                i == 0
                or output[i - 1].get_type() == TokenType.OPERATOR
                or output[i - 1].get_type() == TokenType.LEFT_PARENTHESIS
            ):
                output[i]._number_of_arguments = 1
                output[i]._precedence = 5
                continue

        # return
        return output



"""
    This file is part of the borb (TM) project.
    Copyright (c) 2020-2040 borb (EZ)
    Authors: Joris Schellekens, et al.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3
    as published by the Free Software Foundation with the addition of the
    following permission added to Section 15 as permitted in Section 7(a):
    FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY
    BORB GROUP. BORB GROUP DISCLAIMS THE WARRANTY OF NON INFRINGEMENT
    OF THIRD PARTY RIGHTS

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.

    See the GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program; if not, see http://www.gnu.org/licenses or write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA, 02110-1301 USA.

    The interactive user interfaces in modified source and object code versions
    of this program must display Appropriate Legal Notices, as required under
    Section 5 of the GNU Affero General Public License.
    In accordance with Section 7(b) of the GNU Affero General Public License,
    a covered work must retain the producer line in every PDF that is created
    or manipulated using borb.

    You can be released from the requirements of the license by purchasing
    a commercial license. Buying such a license is mandatory as soon as you
    develop commercial activities involving the borb software without
    disclosing the source code of your own applications.

    These activities include: offering paid services to customers as an ASP,
    serving PDFs on the fly in a web application, shipping borb with a closed
    source product.

    For more information, please contact borb Software Corp. at this
    address: joris.schellekens.1989@gmail.com
"""

# color
from .color.color_extraction import ColorExtraction

# export
from .export.html_to_pdf.html_to_pdf import HTMLToPDF

try:
    from .export.markdown_to_pdf.markdown_to_pdf import MarkdownToPDF
except:
    pass
try:
    from .export.pdf_to_mp3 import PDFToMP3
except:
    pass
from .export.pdf_to_jpg import PDFToJPG
from .export.pdf_to_svg import PDFToSVG

# image
from .image.image_extraction import ImageExtraction
from .image.image_format_optimization import ImageFormatOptimization

# location
from .location.location_filter import LocationFilter

# ocr
try:
    from .ocr.ocr_as_optional_content_group import OCRAsOptionalContentGroup
    from .ocr.ocr_image_render_event_listener import (
        OCREvent,
        OCRImageRenderEventListener,
    )
except:
    pass

# table
from .table.table_detection_by_lines import TableDetectionByLines

# text
from .text.simple_text_extraction import SimpleTextExtraction
from .text.simple_non_ligature_text_extraction import SimpleNonLigatureTextExtraction
from .text.simple_find_replace import SimpleFindReplace

# text (structure)
from .text.simple_paragraph_extraction import SimpleParagraphExtraction
from .text.simple_line_of_text_extraction import SimpleLineOfTextExtraction
from .text.regular_expression_text_extraction import (
    RegularExpressionTextExtraction,
    PDFMatch,
)

# text (filter)
from .text.font_color_filter import FontColorFilter
from .text.font_extraction import FontExtraction
from .text.font_name_filter import FontNameFilter

# text (keywords, NLP)
try:
    from .text.text_rank_keyword_extraction import TextRankKeywordExtraction
except:
    pass
from .text.tf_idf_keyword_extraction import TFIDFKeywordExtraction
from .text.stop_words import FRENCH_STOP_WORDS, ENGLISH_STOP_WORDS



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A parser is a software component that takes input data (frequently text) and builds a data structure
– often some kind of parse tree, abstract syntax tree or other hierarchical structure, giving a structural
representation of the input while checking for correct syntax.
The parsing may be preceded or followed by other steps, or these may be combined into a single step.
"""
import typing

from borb.pdf.canvas.layout.equation.token import Token, TokenType
from borb.pdf.canvas.layout.equation.tokenizer import Tokenizer


class Parser:
    """
    A parser is a software component that takes input data (frequently text) and builds a data structure
    – often some kind of parse tree, abstract syntax tree or other hierarchical structure, giving a structural
    representation of the input while checking for correct syntax.
    The parsing may be preceded or followed by other steps, or these may be combined into a single step.
    """

    @staticmethod
    def _to_postfix(infix_expression: str) -> typing.List[Token]:
        tokens: typing.List[Token] = Tokenizer.tokenize(infix_expression)
        postfix: typing.List[Token] = []
        operators: typing.List[Token] = []
        for i, t in enumerate(tokens):
            if t.get_type() == TokenType.NUMBER:
                postfix += [t]
                continue
            if t.get_type() == TokenType.VARIABLE:
                postfix += [t]
                continue
            if t.get_type() == TokenType.FUNCTION:
                operators += [t]
                continue
            if t.get_type() == TokenType.OPERATOR:
                while (
                    len(operators) > 0
                    and operators[-1].get_type() != TokenType.LEFT_PARENTHESIS
                    and (
                        operators[-1].get_precedence() > t.get_precedence()
                        or (
                            operators[-1].get_precedence() == t.get_precedence()
                            and t.get_is_left_associative()
                        )
                    )
                ):
                    postfix += [operators[-1]]
                    operators.pop(-1)
                operators += [t]
                continue
            if t.get_type() == TokenType.COMMA:
                while (
                    len(operators) > 0
                    and operators[-1].get_type() != TokenType.LEFT_PARENTHESIS
                ):
                    postfix += [operators[-1]]
                    operators.pop(-1)
                continue
            if t.get_type() == TokenType.LEFT_PARENTHESIS:
                operators += [t]
                continue
            if t.get_type() == TokenType.RIGHT_PARENTHESIS:
                assert len(operators) > 0
                while (
                    len(operators) > 0
                    and operators[-1].get_type() != TokenType.LEFT_PARENTHESIS
                ):
                    postfix += [operators[-1]]
                    operators.pop(-1)
                assert len(operators) > 0
                assert operators[-1].get_type() == TokenType.LEFT_PARENTHESIS
                operators.pop(-1)
                if (
                    len(operators) > 0
                    and operators[-1].get_type() == TokenType.FUNCTION
                ):
                    postfix += [operators[-1]]
                    operators.pop(-1)
                continue

        # while there are tokens on the operator stack:
        #   If the operator token on the top of the stack is a parenthesis, then there are mismatched parentheses.
        #   {assert the operator on top of the stack is not a (left) parenthesis}
        #   pop the operator from the operator stack onto the output queue
        while len(operators) > 0:
            postfix += [operators[-1]]
            operators.pop(-1)

        # return
        return postfix

    @staticmethod
    def to_abstract_syntax_tree(s: str) -> Token:
        """
        This function converts a str to an abstract syntax tree
        :param s:   the input str
        :return:    a Token, representing the root of the syntax tree
        """
        args: typing.List[Token] = []
        postfix: typing.List[Token] = Parser._to_postfix(s)
        for i, t in enumerate(postfix):

            if t.get_type() == TokenType.NUMBER:
                args += [t]
                continue

            if t.get_type() == TokenType.VARIABLE:
                args += [t]
                continue

            if t.get_type() == TokenType.OPERATOR:
                assert len(args) >= t.get_number_of_arguments()
                for _ in range(0, t.get_number_of_arguments()):
                    t._children += [args[-1]]
                    args.pop(-1)
                args += [t]
                continue

            if t.get_type() == TokenType.FUNCTION:
                assert len(args) >= t.get_number_of_arguments()
                for _ in range(0, t.get_number_of_arguments()):
                    t._children += [args[-1]]
                    args.pop(-1)
                args += [t]
                continue

        # check
        assert len(args) == 1

        # return
        return args[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A SmartArt graphic is a visual representation of your information and ideas.
You create one by choosing a layout that fits your message.
Some layouts (such as organization charts and Venn diagrams) portray specific kinds of information,
while others simply enhance the appearance of a bulleted list.
"""

import math
import random
import types
import typing
from decimal import Decimal

from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.canvas.layout.table.table import Table, TableCell
from borb.pdf.canvas.layout.table.flexible_column_width_table import (
    FlexibleColumnWidthTable,
)
from borb.pdf.canvas.layout.list.list import List
from borb.pdf.canvas.layout.list.unordered_list import UnorderedList
from borb.pdf.canvas.color.color import HexColor, Color
from borb.pdf.canvas.layout.image.image import Image
from borb.pdf.page.page import Page
from borb.pdf.canvas.layout.shape.connected_shape import ConnectedShape
from borb.pdf.canvas.layout.page_layout.inline_flow import InlineFlow
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement, Alignment
from borb.pdf.canvas.line_art.line_art_factory import LineArtFactory


class SmartArt:
    """
    A SmartArt graphic is a visual representation of your information and ideas.
    You create one by choosing a layout that fits your message.
    Some layouts (such as organization charts and Venn diagrams) portray specific kinds of information,
    while others simply enhance the appearance of a bulleted list.
    """

    @staticmethod
    def alternating_picture_list(
        texts: typing.List[str],
        pictures: typing.List[str],
        font_size: Decimal = Decimal(12),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show a series of pictures from top to bottom. Text appears alternately on the right or left of the picture.
        :param texts:               the typing.List[str] to be used as
        :param pictures             the typing.List[str] to be used as image URLs
        :param font_size:           the font_size to be used
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(texts) > 0
        assert len(pictures) > 0
        assert len(texts) == len(pictures)
        assert font_size > 0

        N: int = len(texts)
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=3, number_of_rows=N
        )

        # add content
        for i in range(0, N):
            if i % 2 == 0:
                table.add(
                    TableCell(
                        Paragraph(
                            texts[i],
                            font_color=font_color,
                            font_size=font_size,
                            background_color=background_color,
                        ),
                        col_span=2,
                        border_left=False,
                        border_top=(True if i != 0 else False),
                        border_bottom=(True if i != N - 1 else False),
                        background_color=background_color,
                        border_color=HexColor("FFFFFF"),
                        border_width=font_size / 2,
                    )
                )
                table.add(
                    TableCell(
                        Image(pictures[i], width=Decimal(64), height=Decimal(64)),
                        border_right=False,
                        border_top=(True if i != 0 else False),
                        border_bottom=(True if i != N - 1 else False),
                        border_color=HexColor("FFFFFF"),
                        background_color=background_color,
                        border_width=font_size / 2,
                    )
                )
            else:
                table.add(
                    TableCell(
                        Image(pictures[i], width=Decimal(64), height=Decimal(64)),
                        border_left=False,
                        border_bottom=(True if i != N - 1 else False),
                        background_color=background_color,
                        border_color=HexColor("FFFFFF"),
                        border_width=font_size / 2,
                    )
                )
                table.add(
                    TableCell(
                        Paragraph(
                            texts[i],
                            font_color=font_color,
                            font_size=font_size,
                            background_color=background_color,
                        ),
                        border_right=False,
                        border_bottom=(True if i != N - 1 else False),
                        col_span=2,
                        background_color=background_color,
                        border_color=HexColor("FFFFFF"),
                        border_width=font_size / 2,
                    )
                )

        table.set_padding_on_all_cells(font_size, font_size, font_size, font_size)

        # return
        return table

    @staticmethod
    def ascending_block_list(
        text_level_1: typing.List[str],
        text_level_2: typing.List[typing.List[str]],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show groups of related ideas or lists of information.
        The text shapes increase in height sequentially, and the Level 1 text displays vertically.
        :param text_level_1:        the typing.List[str] to be used as level 1 text
        :param text_level_2         the typing.List[typing.List[str]] to be used as level 2 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text_level_1) > 0
        assert len(text_level_2) > 0
        assert all([len(x) > 0 for x in text_level_2])
        assert len(text_level_1) == len(text_level_2)
        assert font_size > 0
        N: int = len(text_level_1)
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=N, number_of_rows=2 + (N - 1)
        )
        for i in range(0, N - 1):
            for j in range(0, N):
                if N - j - 1 > i:
                    table.add(
                        TableCell(
                            Paragraph(" "),
                            border_top=False,
                            border_right=False,
                            border_bottom=False,
                            border_left=False,
                        )
                    )
                else:
                    table.add(
                        TableCell(
                            Paragraph(" "),
                            background_color=background_color,
                            border_width=font_size / 2,
                            border_color=HexColor("ffffff"),
                            border_right=True if j != N - 1 else False,
                            border_left=True if j != 0 else False,
                            border_top=False,
                            border_bottom=False,
                        )
                    )

        for i in range(0, N):
            table.add(
                TableCell(
                    Paragraph(
                        text_level_1[i],
                        font_size=font_size + Decimal(2),
                        font="Helvetica-Bold",
                        font_color=font_color,
                        background_color=background_color,
                    ),
                    background_color=background_color,
                    border_width=font_size / 2,
                    border_color=HexColor("ffffff"),
                    border_right=True if i != N - 1 else False,
                    border_left=True if i != 0 else False,
                    border_top=False,
                    border_bottom=False,
                    padding_top=font_size,
                    padding_right=font_size,
                    padding_bottom=font_size,
                    padding_left=font_size,
                )
            )
        for i in range(0, N):
            ul: UnorderedList = UnorderedList()
            for li in text_level_2[i]:
                ul.add(Paragraph(li, font_size=font_size, font_color=font_color))
            table.add(
                TableCell(
                    ul,
                    background_color=background_color,
                    border_width=font_size / 2,
                    border_color=HexColor("ffffff"),
                    border_right=True if i != N - 1 else False,
                    border_left=True if i != 0 else False,
                    border_top=False,
                    border_bottom=False,
                    padding_top=font_size,
                    padding_right=font_size,
                    padding_bottom=font_size,
                    padding_left=font_size,
                )
            )
        return table

    @staticmethod
    def basic_bending_process(
        text: typing.List[str],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show a long or non-linear sequence or steps in a task, process, or workflow.
        Works best with Level 1 text only. Maximizes both horizontal and vertical display space for shapes.
        :param text:                the typing.List[str] to be used as level 1 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text) > 0
        assert font_size > 0
        ncols: int = math.ceil(math.sqrt(len(text)))
        nrows: int = math.ceil(len(text) / ncols)

        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=ncols * 2 - 1, number_of_rows=nrows * 2 - 1
        )

        filling_left_to_right: bool = True
        elements_to_add: typing.List[str] = text
        while len(elements_to_add) > 0:
            next_row: typing.List[typing.Optional[str]] = elements_to_add[0:ncols]  # type: ignore [assignment]
            elements_to_add = elements_to_add[ncols:]
            while len(next_row) != ncols:
                next_row.append(None)
            if not filling_left_to_right:
                next_row.reverse()

            # add <data> + <arrow>
            for i, e in enumerate(next_row):
                if e is None:
                    table.add(Paragraph(" "))
                else:
                    table.add(
                        TableCell(
                            Paragraph(
                                e,
                                font_size=font_size,
                                font_color=font_color,
                                background_color=background_color,
                            ),
                            background_color=background_color,
                        )
                    )

                #
                # ADD ARROW BETWEEN BLOCKS
                #
                add_arrow: bool = False
                add_empty_paragraph_instead_of_arrow: bool = True
                if filling_left_to_right:
                    if (i != len(next_row) - 1) and next_row[i + 1] is not None:
                        add_arrow = True
                if not filling_left_to_right and next_row[i] is not None:
                    add_arrow = True
                if i == len(next_row) - 1:
                    add_arrow = False
                    add_empty_paragraph_instead_of_arrow = False
                if not add_arrow:
                    if add_empty_paragraph_instead_of_arrow:
                        table.add(Paragraph(" "))
                else:
                    # add arrow (depending on the fill direction of this line)
                    if filling_left_to_right:
                        table.add(
                            ConnectedShape(
                                LineArtFactory.arrow_right(
                                    Rectangle(
                                        Decimal(0),
                                        Decimal(0),
                                        Decimal(32),
                                        Decimal(32),
                                    )
                                ),
                                stroke_color=foreground_color,
                                fill_color=foreground_color,
                            )
                        )
                    else:
                        table.add(
                            ConnectedShape(
                                LineArtFactory.arrow_left(
                                    Rectangle(
                                        Decimal(0),
                                        Decimal(0),
                                        Decimal(32),
                                        Decimal(32),
                                    )
                                ),
                                stroke_color=foreground_color,
                                fill_color=foreground_color,
                            )
                        )

            # add <empty> + <arrow down>
            if len(elements_to_add) > 0:

                # add arrow down
                if filling_left_to_right:
                    for _ in range(0, ncols * 2 - 2):
                        table.add(Paragraph(" "))
                    table.add(
                        ConnectedShape(
                            LineArtFactory.arrow_down(
                                Rectangle(
                                    Decimal(0), Decimal(0), Decimal(32), Decimal(32)
                                )
                            ),
                            stroke_color=foreground_color,
                            fill_color=foreground_color,
                        )
                    )
                else:
                    table.add(
                        ConnectedShape(
                            LineArtFactory.arrow_down(
                                Rectangle(
                                    Decimal(0), Decimal(0), Decimal(32), Decimal(32)
                                )
                            ),
                            stroke_color=foreground_color,
                            fill_color=foreground_color,
                        )
                    )
                    for _ in range(0, ncols * 2 - 2):
                        table.add(Paragraph(" "))

            # reverse line direction
            filling_left_to_right = not filling_left_to_right

        table.set_padding_on_all_cells(font_size, font_size, font_size, font_size)
        table.no_borders()

        # return
        return table

    @staticmethod
    def closed_chevron_process(
        text: typing.List[str],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show a progression, a timeline, or sequential steps in a task, process, or workflow, or to emphasize movement or direction.
        Can be used to emphasize information in the starting shape. Works best with Level 1 text only.
        :param text:
        :param font_size:
        :param foreground_color:
        :param background_color:
        :param font_color:
        :return:
        """
        assert font_size > 0
        N = len(text)
        table = FlexibleColumnWidthTable(number_of_columns=N * 2 + 1, number_of_rows=2)

        #
        # 1st row
        #

        # add triangle shape
        Z: Decimal = Decimal(0)
        H: Decimal = Decimal(50)
        table.add(
            TableCell(
                ConnectedShape(
                    [(Z, H), (H, H), (H, Z)],
                    fill_color=foreground_color,
                    stroke_color=foreground_color,
                )
            )
        )

        c0: Color = foreground_color
        c1: Color = background_color
        for i, li in enumerate(text):
            if i % 2 == 0:
                c0 = foreground_color
                c1 = background_color
            else:
                c0 = background_color
                c1 = foreground_color

            # add text
            table.add(
                TableCell(
                    Paragraph(li, font_size=font_size, font_color=font_color),
                    background_color=c0,
                    row_span=2,
                    padding_left=font_size,
                    padding_right=font_size,
                    padding_top=font_size,
                    padding_bottom=font_size,
                )
            )
            if i != N - 1:
                table.add(
                    TableCell(
                        ConnectedShape(
                            [(Z, Z), (Z, H), (H, Z)], fill_color=c0, stroke_color=c0
                        ),
                        background_color=c1,
                    )
                )
            else:
                table.add(
                    TableCell(
                        ConnectedShape(
                            [(Z, Z), (Z, H), (H, Z)], fill_color=c0, stroke_color=c0
                        )
                    )
                )

        #
        # 2nd row
        #

        # add triangle shape
        table.add(
            TableCell(
                ConnectedShape(
                    [(Z, Z), (H, H), (H, Z)],
                    fill_color=foreground_color,
                    stroke_color=foreground_color,
                )
            )
        )

        for i in range(0, N):
            if i % 2 == 0:
                c0 = foreground_color
                c1 = background_color
            else:
                c0 = background_color
                c1 = foreground_color

            # add triangle shape
            if i != N - 1:
                table.add(
                    TableCell(
                        ConnectedShape(
                            [(Z, Z), (Z, H), (H, H)], fill_color=c0, stroke_color=c0
                        ),
                        background_color=c1,
                    )
                )
            else:
                table.add(
                    TableCell(
                        ConnectedShape(
                            [(Z, Z), (Z, H), (H, H)], fill_color=c0, stroke_color=c0
                        )
                    )
                )
        table.no_borders()

        # return
        return table

    @staticmethod
    def descending_block_list(
        text_level_1: typing.List[str],
        text_level_2: typing.List[typing.List[str]],
        font_size: Decimal = Decimal(12),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show groups of related ideas or lists of information.
        The text shapes decrease in height sequentially, and the Level 1 text displays vertically.
        :param text_level_1:        the typing.List[str] to be used as level 1 text
        :param text_level_2         the typing.List[typing.List[str]] to be used as level 2 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text_level_1) > 0
        assert len(text_level_2) > 0
        assert all([len(x) > 0 for x in text_level_2])
        assert len(text_level_1) == len(text_level_2)
        assert font_size > 0
        N: int = len(text_level_1)
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=N, number_of_rows=2 + (N - 1)
        )
        for i in range(0, N - 1):
            for j in range(0, N):
                if j > i:
                    table.add(
                        TableCell(
                            Paragraph(" "),
                            border_top=False,
                            border_right=False,
                            border_bottom=False,
                            border_left=False,
                        )
                    )
                else:
                    table.add(
                        TableCell(
                            Paragraph(" "),
                            background_color=background_color,
                            border_width=font_size / 2,
                            border_color=HexColor("ffffff"),
                            border_right=True if j != N - 1 else False,
                            border_left=True if j != 0 else False,
                            border_top=False,
                            border_bottom=False,
                        )
                    )

        for i in range(0, N):
            table.add(
                TableCell(
                    Paragraph(
                        text_level_1[i],
                        font_size=font_size + Decimal(2),
                        font="Helvetica-Bold",
                        font_color=font_color,
                        background_color=background_color,
                    ),
                    background_color=background_color,
                    border_width=font_size / 2,
                    border_color=HexColor("ffffff"),
                    border_right=True if i != N - 1 else False,
                    border_left=True if i != 0 else False,
                    border_top=False,
                    border_bottom=False,
                    padding_top=font_size,
                    padding_right=font_size,
                    padding_bottom=font_size,
                    padding_left=font_size,
                )
            )
        for i in range(0, N):
            ul: UnorderedList = UnorderedList()
            for li in text_level_2[i]:
                ul.add(Paragraph(li, font_size=font_size, font_color=font_color))
            table.add(
                TableCell(
                    ul,
                    background_color=background_color,
                    border_width=font_size / 2,
                    border_color=HexColor("ffffff"),
                    border_right=True if i != N - 1 else False,
                    border_left=True if i != 0 else False,
                    border_top=False,
                    border_bottom=False,
                    padding_top=font_size,
                    padding_right=font_size,
                    padding_bottom=font_size,
                    padding_left=font_size,
                )
            )
        return table

    @staticmethod
    def horizontal_bullet_list(
        text_level_1: typing.List[str],
        text_level_2: typing.List[typing.List[str]],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        A horizontal bullet list is used to show non-sequential information or grouped data, yet it will not affect the outcome of the other data.
        The first block on the left is related to the second block as well as the third block but the process of the
        second block is not dependent on how the process on the first block is implemented.
        Apart from this, each block of information has the same emphasis, and it does not connote any direction.
        :param text_level_1:        the typing.List[str] to be used as level 1 text
        :param text_level_2         the typing.List[typing.List[str]] to be used as level 2 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text_level_1) > 0
        assert len(text_level_2) > 0
        assert all([len(x) > 0 for x in text_level_2])
        assert len(text_level_1) == len(text_level_2)
        assert font_size > 0
        N: int = len(text_level_1)
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=N, number_of_rows=2
        )

        # add headers
        for i in range(0, N):
            table.add(
                TableCell(
                    Paragraph(
                        text_level_1[i],
                        font_size=font_size + Decimal(2),
                        font="Helvetica-Bold",
                        background_color=background_color,
                        font_color=font_color,
                    ),
                    border_top=False,
                    border_bottom=False,
                    border_left=(True if i != 0 else False),
                    border_right=(True if i != N - 1 else False),
                    border_width=font_size / 2,
                    background_color=background_color,
                    border_color=HexColor("FFFFFF"),
                )
            )

        # add body
        for i in range(0, N):
            ul: UnorderedList = UnorderedList()
            for li in text_level_2[i]:
                ul.add(
                    Paragraph(
                        li,
                        font_size=font_size,
                        background_color=foreground_color,
                        font_color=font_color,
                    )
                )
            table.add(
                TableCell(
                    ul,
                    border_top=False,
                    border_bottom=False,
                    border_left=(True if i != 0 else False),
                    border_right=(True if i != N - 1 else False),
                    border_width=font_size / 2,
                    background_color=foreground_color,
                    border_color=HexColor("FFFFFF"),
                )
            )
        table.set_padding_on_all_cells(font_size, font_size, font_size, font_size)

        # return
        return table

    @staticmethod
    def horizontal_process(
        text: typing.List[str],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show a progression or sequential steps in a task, process, or workflow.
        :param text:                the typing.List[str] to be used as level 1 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text) > 0
        assert font_size > 0
        N: int = len(text)
        n: int = N * 2 - 1
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=n, number_of_rows=1
        )
        for i, e in enumerate(text):
            table.add(
                TableCell(
                    Paragraph(
                        e,
                        font_size=font_size,
                        font_color=font_color,
                        background_color=background_color,
                    ),
                    background_color=background_color,
                )
            )
            if i != len(text) - 1:
                table.add(
                    ConnectedShape(
                        LineArtFactory.arrow_right(
                            Rectangle(
                                Decimal(0),
                                Decimal(0),
                                Decimal(32),
                                Decimal(32),
                            )
                        ),
                        stroke_color=foreground_color,
                        fill_color=foreground_color,
                    )
                )

        table.set_padding_on_all_cells(font_size, font_size, font_size, font_size)
        table.no_borders()

        # return
        return table

    @staticmethod
    def matrix(
        text: typing.List[str],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        A matrix is a rectangular array of elements, arranged in rows and columns, and can be used to show the placement of concepts along two axes.
        For example, you can use a matrix to illustrate the four possible combinations of two concepts or ingredients.
        :param text:                the typing.List[str] to be used in the 4 boxes
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text) == 4
        assert font_size > 0
        modified_table: LayoutElement = (
            FlexibleColumnWidthTable(number_of_columns=2, number_of_rows=2)
            .add(
                TableCell(
                    Paragraph(
                        text[0],
                        font_size=font_size,
                        text_alignment=Alignment.CENTERED,
                        horizontal_alignment=Alignment.CENTERED,
                        font_color=font_color,
                    ),
                    border_width=font_size / 2,
                    border_top=False,
                    border_right=True,
                    border_bottom=True,
                    border_left=False,
                    border_color=background_color,
                )
            )
            .add(
                TableCell(
                    Paragraph(
                        text[1],
                        font_size=font_size,
                        text_alignment=Alignment.CENTERED,
                        horizontal_alignment=Alignment.CENTERED,
                        font_color=font_color,
                    ),
                    border_width=font_size / 2,
                    border_top=False,
                    border_right=False,
                    border_bottom=True,
                    border_left=True,
                    border_color=background_color,
                )
            )
            .add(
                TableCell(
                    Paragraph(
                        text[2],
                        font_size=font_size,
                        text_alignment=Alignment.CENTERED,
                        horizontal_alignment=Alignment.CENTERED,
                        font_color=font_color,
                    ),
                    border_width=font_size / 2,
                    border_top=True,
                    border_right=True,
                    border_bottom=False,
                    border_left=False,
                    border_color=background_color,
                )
            )
            .add(
                TableCell(
                    Paragraph(
                        text[3],
                        font_size=font_size,
                        text_alignment=Alignment.CENTERED,
                        horizontal_alignment=Alignment.CENTERED,
                        font_color=font_color,
                    ),
                    border_width=font_size / 2,
                    border_top=True,
                    border_right=False,
                    border_bottom=False,
                    border_left=True,
                    border_color=background_color,
                )
            )
            .set_padding_on_all_cells(font_size, font_size, font_size, font_size)
            .set_background_color_on_all_cells(foreground_color)
        )

        # get the paint method
        prev_paint = modified_table.paint

        # build a modified paint method
        def _modified_paint_method(obj, page: Page, available_space: Rectangle):
            lbox: Rectangle = obj.get_layout_box(available_space)
            lbox = lbox.grow(font_size)
            ConnectedShape(
                LineArtFactory.diamond(lbox),
                stroke_color=background_color,
                fill_color=background_color,
            ).paint(page, lbox)
            prev_paint(page, available_space)

        # assign
        modified_table.paint = types.MethodType(_modified_paint_method, modified_table)  # type: ignore [assignment]

        # return
        output_element: LayoutElement = InlineFlow().add(modified_table)
        output_element._margin_top = font_size
        output_element._margin_right = font_size
        output_element._margin_bottom = font_size
        output_element._margin_left = font_size
        return output_element

    @staticmethod
    def opposing_ideas(
        text: typing.List[str],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show two opposing or contrasting ideas. Has two Level 1 items.
        Works well with large amounts of text.
        :param text:                the typing.List[str] to be used as level 1 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text) > 0
        assert font_size > 0
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=4, number_of_rows=3
        )

        # row 1
        table.add(Paragraph(" "))
        table.add(
            TableCell(
                Paragraph(
                    text[0],
                    font_size=font_size,
                    font_color=font_color,
                    background_color=background_color,
                ),
                row_span=3,
                border_radius_bottom_left=Decimal(20),
                background_color=background_color,
            )
        )
        table.add(
            TableCell(
                Paragraph(
                    text[1],
                    font_size=font_size,
                    font_color=font_color,
                    background_color=foreground_color,
                ),
                row_span=3,
                border_radius_top_right=Decimal(20),
                background_color=foreground_color,
            )
        )
        table.add(Paragraph(" "))

        # row 2
        table.add(
            ConnectedShape(
                LineArtFactory.arrow_up(
                    Rectangle(Decimal(0), Decimal(0), Decimal(32), Decimal(32))
                ),
                stroke_color=background_color,
                fill_color=background_color,
            )
        )
        table.add(
            ConnectedShape(
                LineArtFactory.arrow_down(
                    Rectangle(Decimal(0), Decimal(0), Decimal(32), Decimal(32))
                ),
                stroke_color=foreground_color,
                fill_color=foreground_color,
            )
        )

        # row 3
        table.add(Paragraph(" "))
        table.add(Paragraph(" "))

        # set global properties
        table.set_padding_on_all_cells(font_size, font_size, font_size, font_size)
        table.no_borders()

        # return
        return table

    @staticmethod
    def picture_list(
        texts: typing.List[str],
        pictures: typing.List[str],
        font_size: Decimal = Decimal(12),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show a series of pictures from top to bottom. Text appears on the right of the picture.
        :param texts:               the typing.List[str] to be used as
        :param pictures             the typing.List[str] to be used as image URLs
        :param font_size:           the font_size to be used
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(texts) > 0
        assert len(pictures) > 0
        assert len(texts) == len(pictures)
        assert font_size > 0

        N: int = len(texts)
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=3, number_of_rows=N
        )

        # add content
        for i in range(0, N):
            table.add(
                TableCell(
                    Image(pictures[i], width=Decimal(64), height=Decimal(64)),
                    border_right=True,
                    border_left=False,
                    border_top=(True if i != 0 else False),
                    border_bottom=(True if i != N - 1 else False),
                    border_color=HexColor("FFFFFF"),
                    background_color=background_color,
                    border_width=font_size / 2,
                )
            )
            table.add(
                TableCell(
                    Paragraph(
                        texts[i],
                        font_color=font_color,
                        font_size=font_size,
                        background_color=background_color,
                    ),
                    col_span=2,
                    border_right=False,
                    border_left=True,
                    border_top=(True if i != 0 else False),
                    border_bottom=(True if i != N - 1 else False),
                    background_color=background_color,
                    border_color=HexColor("FFFFFF"),
                    border_width=font_size / 2,
                )
            )

        table.set_padding_on_all_cells(font_size, font_size, font_size, font_size)

        # return
        return table

    @staticmethod
    def table_hierarcy(
        text: typing.Any,
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show groups of information built from top to bottom, and the hierarchies within each group.
        This layout does not contain connecting lines.
        :param text:                the typing.List[str] to be used as level 1 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """

        def _cols(e) -> int:
            if isinstance(e, str) or e[1] is None:
                return 1
            return sum([_cols(x) for x in e[1]])

        def _rows(e) -> int:
            if isinstance(e, str) or e[1] is None:
                return 1
            return 1 + max([_rows(x) for x in e[1]])

        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=_cols(text), number_of_rows=_rows(text)
        )
        row: typing.List[typing.Any] = [text]
        number_of_rows: int = _rows(text)
        j: int = 0
        while len(row) > 0:
            next_row: typing.List[typing.Any] = []
            for i, e in enumerate(row):
                c: Color = background_color
                if random.randint(0, 100) > 30:
                    c = foreground_color
                key: typing.Optional[str] = None
                children: typing.List[typing.Any] = []
                if isinstance(e, str):
                    key = e
                if isinstance(e, tuple):
                    key = e[0]
                    children.extend(e[1])
                table.add(
                    TableCell(
                        Paragraph(
                            key or "",
                            text_alignment=Alignment.CENTERED,
                            horizontal_alignment=Alignment.CENTERED,
                            font_size=font_size,
                            font_color=font_color,
                            background_color=c,
                        ),
                        border_left=True if i != 0 else False,
                        border_right=(True if i != len(row) - 1 else False),
                        border_top=False if j == 0 else True,
                        border_bottom=False if j == number_of_rows - 1 else True,
                        border_width=font_size / 2,
                        border_color=HexColor("ffffff"),
                        col_span=_cols(e),
                        background_color=c,
                    )
                )
                next_row.extend(children)
            row = next_row
            j += 1

        table.set_padding_on_all_cells(font_size, font_size, font_size, font_size)

        # table
        return table

    @staticmethod
    def vertical_bullet_list(
        text_level_1: typing.List[str],
        text_level_2: typing.List[typing.List[str]],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        A vertical bullet list is used to show non-sequential information or grouped data, yet it will not affect the outcome of the other data.
        The first block on the top is related to the second block as well as the third block but the process of the
        second block is not dependent on how the process on the first block is implemented.
        Apart from this, each block of information has the same emphasis, and it does not connote any direction.
        :param text_level_1:        the typing.List[str] to be used as level 1 text
        :param text_level_2         the typing.List[typing.List[str]] to be used as level 2 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text_level_1) > 0
        assert len(text_level_2) > 0
        assert all([len(x) > 0 for x in text_level_2])
        assert len(text_level_1) == len(text_level_2)
        assert font_size > 0
        N: int = len(text_level_1)
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_columns=1, number_of_rows=N * 2
        )
        for i in range(0, N):
            table.add(
                TableCell(
                    Paragraph(
                        text_level_1[i],
                        font_size=font_size + Decimal(2),
                        font="Helvetica-Bold",
                        font_color=font_color,
                        background_color=background_color,
                    ),
                    background_color=background_color,
                    border_radius_top_right=Decimal(20),
                    border_radius_bottom_right=Decimal(20),
                    border_radius_bottom_left=Decimal(20),
                    border_radius_top_left=Decimal(20),
                    border_color=background_color,
                    padding_top=font_size,
                    padding_right=font_size,
                    padding_bottom=font_size,
                    padding_left=font_size,
                )
            )
            ul: UnorderedList = UnorderedList()
            for li in text_level_2[i]:
                ul.add(Paragraph(li, font_size=font_size))
            table.add(
                TableCell(
                    ul,
                    border_top=False,
                    border_right=False,
                    border_bottom=False,
                    border_left=False,
                    padding_top=font_size,
                    padding_right=font_size,
                    padding_bottom=font_size,
                    padding_left=font_size,
                )
            )

        # return
        return table

    @staticmethod
    def vertical_process(
        text: typing.List[str],
        font_size: Decimal = Decimal(12),
        foreground_color: Color = HexColor("#6CAE75"),
        background_color: Color = HexColor("#8BBD8B"),
        font_color: Color = HexColor("#FFFFFF"),
    ) -> LayoutElement:
        """
        Use to show a progression or sequential steps in a task, process, or workflow.
        :param text:                the typing.List[str] to be used as level 1 text
        :param font_size:           the font_size to be used
        :param foreground_color:    the Color to be used in the foreground
        :param background_color:    the color to be used in the background
        :param font_color:          the Color to be used for text
        :return:                    a LayoutElement
        """
        assert len(text) > 0
        assert font_size > 0
        N: int = len(text)
        n: int = N * 2 - 1
        table: FlexibleColumnWidthTable = FlexibleColumnWidthTable(
            number_of_rows=n, number_of_columns=1
        )
        for i, e in enumerate(text):
            table.add(
                TableCell(
                    Paragraph(
                        e,
                        font_size=font_size,
                        font_color=font_color,
                        background_color=background_color,
                    ),
                    background_color=background_color,
                )
            )
            if i != len(text) - 1:
                table.add(
                    ConnectedShape(
                        LineArtFactory.arrow_down(
                            Rectangle(
                                Decimal(0),
                                Decimal(0),
                                Decimal(32),
                                Decimal(32),
                            )
                        ),
                        stroke_color=foreground_color,
                        fill_color=foreground_color,
                    )
                )

        table.set_padding_on_all_cells(font_size, font_size, font_size, font_size)
        table.no_borders()

        # return
        return table



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of EventListener extracts all Image objects on a Page
"""
import io
import typing
from typing import List

from PIL import Image as PILImage  # type: ignore [import]

from borb.pdf.document.document import Document
from borb.pdf.canvas.canvas import Canvas
from borb.pdf.canvas.canvas_stream_processor import CanvasStreamProcessor
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.page.page import Page


class ImageExtraction(EventListener):
    """
    This implementation of EventListener extracts all Image objects on a Page
    """

    #
    # CONSTRUCTOR
    #

    def __init__(self):
        """
        Constructs a new SimpleImageExtraction
        """
        self._image_render_info_per_page = {}
        self._current_page: int = -1

    #
    # PRIVATE
    #

    def _begin_page(self, page: Page):
        self._current_page += 1

    def _event_occurred(self, event: "Event") -> None:
        if isinstance(event, BeginPageEvent):
            self._begin_page(event.get_page())
        if isinstance(event, ImageRenderEvent):
            self._render_image(event)

    def _render_image(self, image_render_event: "ImageRenderEvent"):

        # init if needed
        if self._current_page not in self._image_render_info_per_page:
            self._image_render_info_per_page[self._current_page] = []

        # append ImageRenderEvent
        self._image_render_info_per_page[self._current_page].append(
            image_render_event.get_image()
        )

    #
    # PUBLIC
    #

    def get_images(self) -> typing.Dict[int, List[PILImage.Image]]:
        """
        This function returns a typing.List[Image] on a given page
        """
        return self._image_render_info_per_page

    @staticmethod
    def get_images_from_pdf(
        pdf: Document,
    ) -> typing.Dict[int, typing.List[PILImage.Image]]:
        """
        This function returns the images used in a given PDF
        :param pdf:     the PDF to be analysed
        :return:        the images (typing.List[PILImage.Image]) in the PDF
        """
        images_of_each_page: typing.Dict[int, typing.List[PILImage.Image]] = {}
        number_of_pages: int = int(pdf.get_document_info().get_number_of_pages() or 0)
        for page_nr in range(0, number_of_pages):
            # get Page object
            page: Page = pdf.get_page(page_nr)
            page_source: io.BytesIO = io.BytesIO(page["Contents"]["DecodedBytes"])

            # register EventListener
            cse: "ImageExtraction" = ImageExtraction()

            # process Page
            cse._event_occurred(BeginPageEvent(page))
            CanvasStreamProcessor(page, Canvas(), []).read(page_source, [cse])
            cse._event_occurred(EndPageEvent(page))

            # set in page
            images_of_each_page[page_nr] = cse.get_images()[0]

        # return
        return images_of_each_page



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of Event is triggered when an BT instruction is being processed.
"""
from borb.pdf.canvas.event.event_listener import Event


class BeginTextEvent(Event):
    """
    This implementation of Event is triggered when an BT instruction is being processed.
    """

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    #
    # PUBLIC
    #

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of FormField represents a text field.
"""
import typing
import zlib
from decimal import Decimal

from borb.io.read.pdf_object import PDFObject
from borb.io.read.types import Boolean, AnyPDFType
from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, List, Name, Stream, String
from borb.pdf import Alignment
from borb.pdf.canvas.color.color import Color, HexColor, RGBColor
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.forms.form_field import FormField
from borb.pdf.page.page import Page


class TextField(FormField):
    """
    This implementation of FormField represents a text field.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        background_color: typing.Optional[Color] = None,
        border_bottom: bool = True,
        border_color: Color = HexColor("808080"),
        border_left: bool = True,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = True,
        border_top: bool = True,
        border_width: Decimal = Decimal(1),
        default_value: str = "",
        field_name: typing.Optional[str] = None,
        font_color: Color = HexColor("000000"),
        font_size: typing.Optional[Decimal] = Decimal(12),
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: typing.Optional[Decimal] = Decimal(0),
        margin_left: typing.Optional[Decimal] = Decimal(0),
        margin_right: typing.Optional[Decimal] = Decimal(0),
        margin_top: typing.Optional[Decimal] = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        value: str = "",
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        super(TextField, self).__init__(
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            font_size=font_size,
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )
        self._font_color = font_color
        self._value: str = value
        self._default_value: str = default_value
        self._field_name: typing.Optional[str] = field_name
        self._widget_dictionary: typing.Optional[Dictionary] = None

    #
    # PRIVATE
    #

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        # determine layout rectangle
        assert self._font_size is not None
        line_height: Decimal = self._font_size * Decimal(1.2)
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - line_height,
            max(available_space.get_width(), Decimal(64)),
            line_height,
        )

    def _init_widget_dictionary(self, page: Page, layout_box: Rectangle) -> None:

        if self._widget_dictionary is not None:
            return

        root: typing.Optional[PDFObject] = page.get_root()
        assert root is not None
        assert isinstance(root, Dictionary)
        if "XRef" not in root:
            return

        # init page and font resources
        assert self._font_size is not None
        font_resource_name: Name = self._get_font_resource_name(
            StandardType1Font("Helvetica"), page
        )

        # widget resource dictionary
        widget_resources: Dictionary = Dictionary()
        widget_resources.set_is_unique(True)
        widget_resources[Name("Font")] = page["Resources"]["Font"]

        # widget normal appearance
        # fmt: off
        widget_normal_appearance: Stream = Stream()
        widget_normal_appearance.set_is_unique(True)
        widget_normal_appearance[Name("Type")] = Name("XObject")
        widget_normal_appearance[Name("Subtype")] = Name("Form")
        widget_normal_appearance[Name("BBox")] = List()
        widget_normal_appearance[Name("BBox")].set_is_inline(True)  # type: ignore [attr-defined]
        widget_normal_appearance["BBox"].append(bDecimal(0))
        widget_normal_appearance["BBox"].append(bDecimal(0))
        widget_normal_appearance["BBox"].append(bDecimal(layout_box.width))
        widget_normal_appearance["BBox"].append(bDecimal(self._font_size))
        widget_normal_appearance[Name("Resources")] = widget_resources
        widget_normal_appearance[Name("DecodedBytes")] = b"/Tx BMC EMC"
        widget_normal_appearance[Name("Bytes")] = zlib.compress(widget_normal_appearance[Name("DecodedBytes")], 9)
        widget_normal_appearance[Name("Filter")] = Name("FlateDecode")
        widget_normal_appearance[Name("Length")] = bDecimal(len(widget_normal_appearance[Name("Bytes")]))
        # fmt: on

        # widget appearance dictionary
        # fmt: off
        widget_appearance_dictionary: Dictionary = Dictionary()
        widget_appearance_dictionary.set_is_unique(True)
        widget_appearance_dictionary[Name("N")] = widget_normal_appearance
        # fmt: on

        # get Catalog
        catalog: Dictionary = root["XRef"]["Trailer"]["Root"]  # type: ignore [attr-defined]

        # widget dictionary
        # fmt: off
        self._widget_dictionary = Dictionary()
        self._widget_dictionary.set_is_unique(True)
        self._widget_dictionary[Name("Type")] = Name("Annot")
        self._widget_dictionary[Name("Subtype")] = Name("Widget")
        self._widget_dictionary[Name("F")] = bDecimal(4)
        self._widget_dictionary[Name("Rect")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self._widget_dictionary["Rect"].append(bDecimal(layout_box.x))
        self._widget_dictionary["Rect"].append(bDecimal(layout_box.y + layout_box.height - self._font_size))
        self._widget_dictionary["Rect"].append(bDecimal(layout_box.x + layout_box.width))
        self._widget_dictionary["Rect"].append(bDecimal(layout_box.y + layout_box.height))
        self._widget_dictionary[Name("FT")] = Name("Tx")
        self._widget_dictionary[Name("P")] = catalog
        self._widget_dictionary[Name("T")] = String(self._field_name or self._get_auto_generated_field_name(page))
        self._widget_dictionary[Name("V")] = String(self._value)
        self._widget_dictionary[Name("DV")] = String(self._default_value)
        self._widget_dictionary[Name("DR")] = widget_resources
        # fmt: on

        # rendering instructions
        font_color_rgb: RGBColor = self._font_color.to_rgb()
        self._widget_dictionary[Name("DA")] = String(
            "%f %f %f rg /%s %f Tf"
            % (
                float(font_color_rgb.red),
                float(font_color_rgb.green),
                float(font_color_rgb.blue),
                font_resource_name,
                float(self._font_size),
            )
        )
        self._widget_dictionary[Name("AP")] = widget_appearance_dictionary

        # append field to page /Annots
        if "Annots" not in page:
            page[Name("Annots")] = List()
        page["Annots"].append(self._widget_dictionary)

        # append field to catalog
        if "AcroForm" not in catalog:
            catalog[Name("AcroForm")] = Dictionary()
            catalog["AcroForm"][Name("Fields")] = List()
            catalog["AcroForm"][Name("DR")] = widget_resources
            catalog["AcroForm"][Name("NeedAppearances")] = Boolean(True)
        catalog["AcroForm"]["Fields"].append(self._widget_dictionary)

    def _paint_content_box(self, page: "Page", available_space: Rectangle) -> None:

        # determine layout rectangle
        cbox: Rectangle = self._get_content_box(available_space)

        # init self._widget_dictionary
        self._init_widget_dictionary(page, cbox)

        # set location
        # fmt: off
        assert self._font_size is not None
        if self._widget_dictionary is not None:
            self._widget_dictionary["AP"]["N"]["BBox"][2] = bDecimal(cbox.get_width())
            self._widget_dictionary["AP"]["N"]["BBox"][3] = bDecimal(self._font_size)
            self._widget_dictionary["Rect"][0] = bDecimal(cbox.get_x())
            self._widget_dictionary["Rect"][1] = bDecimal(cbox.get_y())
            self._widget_dictionary["Rect"][2] = bDecimal(cbox.get_x() + cbox.get_width())
            self._widget_dictionary["Rect"][3] = bDecimal(cbox.get_y() + cbox.get_height())
        # fmt: on

    #
    # PUBLIC
    #



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
In mathematics, an equation is a formula that expresses the equality of two expressions,
by connecting them with the equals sign =. The word equation and its cognates in other languages may have subtly
different meanings; for example, in French an équation is defined as containing one or more variables,
while in English, any well-formed formula consisting of two expressions related with an equals sign is an equation.
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.layout.table.table import Table, TableCell
from borb.pdf.canvas.layout.table.flexible_column_width_table import (
    FlexibleColumnWidthTable,
)
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.layout.page_layout.inline_flow import InlineFlow
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.layout.equation.parser import Parser
from borb.pdf.canvas.layout.equation.token import Token, TokenType
from borb.pdf.canvas.layout.layout_element import LayoutElement, Alignment


class Equation(InlineFlow):
    """
    In mathematics, an equation is a formula that expresses the equality of two expressions,
    by connecting them with the equals sign =. The word equation and its cognates in other languages may have subtly
    different meanings; for example, in French an équation is defined as containing one or more variables,
    while in English, any well-formed formula consisting of two expressions related with an equals sign is an equation.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        text: str,
        font: typing.Union[Font, str] = "Helvetica",
        font_size: Decimal = Decimal(12),
        vertical_alignment: Alignment = Alignment.TOP,
        horizontal_alignment: Alignment = Alignment.LEFT,
        font_color: Color = HexColor("000000"),
        border_top: bool = False,
        border_right: bool = False,
        border_bottom: bool = False,
        border_left: bool = False,
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_bottom_left: Decimal = Decimal(0),
        border_color: Color = HexColor("000000"),
        border_width: Decimal = Decimal(1),
        padding_top: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        margin_top: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        background_color: typing.Optional[Color] = None,
    ):
        super(Equation, self).__init__()
        self._font = font
        self._font_size = font_size
        self._vertical_alignment = vertical_alignment
        self._horizontal_alignment = horizontal_alignment
        self._font_color = font_color
        self._border_top = border_top
        self._border_right = border_right
        self._border_bottom = border_bottom
        self._border_left = border_left
        self._border_radius_top_left = border_radius_top_left
        self._border_radius_top_right = border_radius_top_right
        self._border_radius_bottom_right = border_radius_bottom_right
        self._border_radius_bottom_left = border_radius_bottom_left
        self._border_color = border_color
        self._border_width = border_width
        self._padding_top = padding_top
        self._padding_right = padding_right
        self._padding_bottom = padding_bottom
        self._padding_left = padding_left
        self._margin_top = margin_top
        self._margin_right = margin_right
        self._margin_bottom = margin_bottom
        self._margin_left = margin_left
        self._background_color = background_color
        self.add(
            self._build(
                Parser.to_abstract_syntax_tree(text),
                font_color=font_color,
                font_size=font_size,
            )
        )

    #
    # PRIVATE
    #

    def _build(
        self,
        t: Token,
        font_color: Color = HexColor("#000000"),
        font_size: Decimal = Decimal(12),
    ) -> LayoutElement:
        if t.get_type() == TokenType.COMMA:
            return ChunkOfText(",", font_size=font_size, font_color=font_color)
        if t.get_type() == TokenType.FUNCTION:
            #
            # abs
            #
            if t.get_text() == "abs":
                c0 = self._build(
                    t.get_children()[0], font_size=font_size, font_color=font_color
                )
                c0._border_width = Decimal(0.75)
                c0._border_left = True
                c0._border_right = True
                return c0
            #
            # any other FUNCTION
            #
            c0: LayoutElement = self._build(
                t.get_children()[0], font_size=font_size, font_color=font_color
            )
            c0._vertical_alignment = Alignment.MIDDLE
            out: Table = FlexibleColumnWidthTable(number_of_columns=4, number_of_rows=1)
            out.add(
                ChunkOfText(
                    t.get_text(),
                    font_size=font_size,
                    font_color=font_color,
                    vertical_alignment=Alignment.MIDDLE,
                )
            )
            out.add(
                ChunkOfText(
                    "(",
                    font_size=font_size,
                    font_color=font_color,
                    vertical_alignment=Alignment.MIDDLE,
                )
            )
            out.add(c0)
            out.add(
                ChunkOfText(
                    ")",
                    font_size=font_size,
                    font_color=font_color,
                    vertical_alignment=Alignment.MIDDLE,
                )
            )
            out.no_borders()
            return out
        if t.get_type() == TokenType.LEFT_PARENTHESIS:
            return ChunkOfText("(", font_size=font_size, font_color=font_color)
        if t.get_type() == TokenType.NUMBER:
            return ChunkOfText(t.get_text(), font_size=font_size, font_color=font_color)
        if t.get_type() == TokenType.OPERATOR:
            #
            # unary minus
            #
            if t.get_text() == "-" and t.get_number_of_arguments() == 1:
                out: Table = FlexibleColumnWidthTable(
                    number_of_columns=2, number_of_rows=1
                )
                out.add(
                    ChunkOfText(
                        "-",
                        font_size=font_size,
                        font_color=font_color,
                        vertical_alignment=Alignment.MIDDLE,
                    )
                )
                out.add(
                    self._build(
                        t.get_children()[0], font_size=font_size, font_color=font_color
                    )
                )
                out.no_borders()
                return out
            #
            # division
            #
            if t.get_text() == "/":
                e0: LayoutElement = self._build(
                    t.get_children()[0], font_size=font_size, font_color=font_color
                )
                e0._horizontal_alignment = Alignment.CENTERED
                e1: LayoutElement = self._build(
                    t.get_children()[1], font_size=font_size, font_color=font_color
                )
                e1._horizontal_alignment = Alignment.CENTERED
                out: Table = FlexibleColumnWidthTable(
                    number_of_columns=1, number_of_rows=2
                )
                out.add(
                    TableCell(
                        e1,
                        border_top=False,
                        border_right=False,
                        border_bottom=True,
                        border_left=False,
                        border_color=font_color,
                        border_width=Decimal(0.75),
                    )
                )
                out.add(
                    TableCell(
                        e0,
                        border_top=False,
                        border_right=False,
                        border_bottom=False,
                        border_left=False,
                        padding_top=Decimal(3),
                    )
                )
                return out
            #
            # power
            #
            if t.get_text() == "^":
                e0: LayoutElement = self._build(
                    t.get_children()[0],
                    font_size=font_size * Decimal(0.5),
                    font_color=font_color,
                )
                e0._horizontal_alignment = Alignment.BOTTOM
                e1: LayoutElement = self._build(
                    t.get_children()[1], font_size=font_size, font_color=font_color
                )
                e1._horizontal_alignment = Alignment.CENTERED
                out: Table = FlexibleColumnWidthTable(
                    number_of_columns=2, number_of_rows=2
                )
                out.add(Paragraph(" ", font_size=font_size * Decimal(0.5)))
                out.add(e0)
                out.add(e1)
                out.add(Paragraph(" "))
                out.set_border_width_on_all_cells(Decimal(0))
                out.no_borders()
                return out
            #
            # any other OPERATOR
            #
            op_text: str = t.get_text()
            if op_text == "*":
                op_text = "·"
            c0: LayoutElement = self._build(
                t.get_children()[0], font_size=font_size, font_color=font_color
            )
            c0._vertical_alignment = Alignment.MIDDLE
            c1: LayoutElement = self._build(
                t.get_children()[1], font_size=font_size, font_color=font_color
            )
            c1._vertical_alignment = Alignment.MIDDLE
            out: Table = FlexibleColumnWidthTable(number_of_columns=3, number_of_rows=1)
            out.add(c1)
            out.add(
                ChunkOfText(
                    op_text,
                    font_size=font_size,
                    font_color=font_color,
                    vertical_alignment=Alignment.MIDDLE,
                    padding_right=Decimal(2),
                    padding_left=Decimal(2),
                )
            )
            out.add(c0)
            out.no_borders()
            return out

        if t.get_type() == TokenType.RIGHT_PARENTHESIS:
            return ChunkOfText(")", font_size=font_size, font_color=font_color)
        if t.get_type() == TokenType.VARIABLE:
            return ChunkOfText(t.get_text(), font_size=font_size, font_color=font_color)

    #
    # PUBLIC
    #



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This implementation of PageLayout adds left/right/top/bottom margins to a Page
and lays out the content on the Page as if there were was a single column to flow text, images, etc into.
Once this column is full, the next page is automatically created.
"""

import copy
import typing
from decimal import Decimal

from borb.pdf.canvas.layout.page_layout.multi_column_layout import SingleColumnLayout
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement


class SingleColumnLayoutWithOverflow(SingleColumnLayout):
    """
    This implementation of PageLayout adds left/right/top/bottom margins to a Page
    and lays out the content on the Page as if there were was a single column to flow text, images, etc into.
    Once this column is full, the next page is automatically created.
    """

    #
    # CONSTRUCTOR
    #

    def __init__(
        self,
        page: "Page",  # type: ignore [name-defined]
        horizontal_margin: typing.Optional[Decimal] = None,
        vertical_margin: typing.Optional[Decimal] = None,
    ):
        super(SingleColumnLayoutWithOverflow, self).__init__(
            page, horizontal_margin, vertical_margin
        )

    #
    # PRIVATE
    #

    @staticmethod
    def _prepare_table_for_relayout(layout_element: LayoutElement):
        from borb.pdf import Table

        assert isinstance(layout_element, Table)
        layout_element._previous_layout_box = None
        layout_element._previous_paint_box = None
        for tc in layout_element._content:
            tc._previous_layout_box = None
            tc._previous_paint_box = None
            tc._forced_layout_box = None
            tc._layout_element._previous_layout_box = None
            tc._layout_element._previous_paint_box = None

    def _split_table(
        self, layout_element: LayoutElement, available_height: Decimal
    ) -> typing.List[LayoutElement]:

        # find out at which row we ought to split the Table
        from borb.pdf import Table

        assert isinstance(layout_element, Table)
        best_row_for_split: typing.Optional[int] = None
        for i in range(0, layout_element._number_of_rows):
            if any([x._row_span != 1 for x in layout_element._get_cells_at_row(i)]):
                continue
            prev_layout_box: typing.Optional[
                Rectangle
            ] = layout_element._get_cells_at_row(i)[0].get_previous_layout_box()
            assert prev_layout_box is not None
            y: Decimal = prev_layout_box.get_y()
            if y < 0:
                continue
            if y < available_height:
                best_row_for_split = i

        # unable to split
        if best_row_for_split is None:
            assert False, (
                "%s is too tall to fit inside column / page."
                % layout_element.__class__.__name__
            )

        # first half of split
        t0 = copy.deepcopy(layout_element)
        t0._number_of_rows = best_row_for_split + 1
        t0._content = [
            x
            for x in t0._content
            if all([y[0] <= best_row_for_split for y in x._table_coordinates])
        ]
        SingleColumnLayoutWithOverflow._prepare_table_for_relayout(t0)

        # second half of split
        t1 = copy.deepcopy(layout_element)
        t1._number_of_rows = layout_element._number_of_rows - best_row_for_split - 1
        t1._content = [
            x
            for x in t1._content
            if all([y[0] > best_row_for_split for y in x._table_coordinates])
        ]
        for tc in t1._content:
            tc._table_coordinates = [
                (y - best_row_for_split - 1, x) for y, x in tc._table_coordinates
            ]
        SingleColumnLayoutWithOverflow._prepare_table_for_relayout(t1)

        # return
        return [t0, t1]

    def _split_blockflow(
        self, layout_element: LayoutElement, available_height: Decimal
    ) -> typing.List[LayoutElement]:
        from borb.pdf import BlockFlow

        assert isinstance(layout_element, BlockFlow)
        return layout_element._content

    #
    # PUBLIC
    #

    def add(self, layout_element: LayoutElement) -> "PageLayout":  # type: ignore [name-defined]
        """
        This method adds a `LayoutElement` to the current `Page`.
        """

        # anything that isn't a Table gets added as expected
        if layout_element.__class__.__name__ not in [
            "BlockFlow",
            "FlexibleColumnWidthTable",
            "FixedColumnWidthTable",
        ]:
            return super(SingleColumnLayout, self).add(layout_element)

        # previous element is used to determine the paragraph spacing
        assert self._page_height is not None
        assert self._page_width is not None
        previous_element_margin_bottom: Decimal = Decimal(0)
        previous_element_y = self._page_height - self._vertical_margin_top
        if self._previous_element is not None:
            previous_element_y = (
                self._previous_element.get_previous_layout_box().get_y()
            )
            previous_element_margin_bottom = self._previous_element.get_margin_bottom()

        # calculate next available height
        available_height: Decimal = (
            previous_element_y
            - self._vertical_margin_bottom
            - self._get_margin_between_elements(self._previous_element, layout_element)
            - max(previous_element_margin_bottom, layout_element.get_margin_top())
            - layout_element.get_margin_bottom()
        )

        # switch to new column if needed
        assert self._page_height
        if available_height < 0:
            self.switch_to_next_column()
            return self.add(layout_element)

        # ask LayoutElement to fit
        lbox: Rectangle = layout_element.get_layout_box(
            Rectangle(
                self._horizontal_margin + layout_element.get_margin_left(),
                Decimal(0),
                self._column_width
                - layout_element.get_margin_right()
                - layout_element.get_margin_left(),
                available_height,
            )
        )
        if lbox.get_height() <= available_height:
            return super(SingleColumnLayout, self).add(layout_element)

        # split Table
        if layout_element.__class__.__name__ in [
            "FlexibleColumnWidthTable",
            "FixedColumnWidthTable",
        ]:
            for t in self._split_table(layout_element, available_height):
                super(SingleColumnLayoutWithOverflow, self).add(t)

        # split BlockFlow
        if layout_element.__class__.__name__ in [
            "BlockFlow",
        ]:
            for t in self._split_blockflow(layout_element, available_height):
                super(SingleColumnLayoutWithOverflow, self).add(t)

        # return
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class provides functions to extract a color palette (typing.List[Color]) from an Image
(specified by str, Path, borb Image, PIL Image)
"""
import math
import typing
from decimal import Decimal
from pathlib import Path

from PIL.Image import Image as pImage
from borb.pdf import Image as bImage, RGBColor, HexColor


class ColorPaletteFromImage:
    """
    This class provides functions to extract a color palette (typing.List[Color]) from an Image
    (specified by str, Path, borb Image, PIL Image)
    """

    @staticmethod
    def _dist(c0: str, c1: str) -> float:
        r_delta = (int(c0[1:3], 16) - int(c1[1:3], 16)) ** 2 / (255**2)
        g_delta = (int(c0[3:5], 16) - int(c1[3:5], 16)) ** 2 / (255**2)
        b_delta = (int(c0[5:7], 16) - int(c1[5:7], 16)) ** 2 / (255**2)
        return math.sqrt(r_delta + g_delta + b_delta) / math.sqrt(3)

    @staticmethod
    def color_palette_from_image(
        img: typing.Union[str, Path, bImage, pImage], limit: int = 4
    ):
        """
        This function calculates the color palette from an Image
        :param img:     the input Image (specified by str, Path, borb Image, or PIL Image)
        :param limit:   the maximum number of colors in the palette
        :return:        a color palette (represented by typing.List[Color])
        """

        # convert everything to bImage
        if isinstance(img, str):
            img = bImage(img)
        if isinstance(img, Path):
            img = bImage(img)
        if isinstance(img, pImage):
            img = bImage(img)
        assert isinstance(img, bImage)
        img.force_load_image()

        # resize
        assert isinstance(img._image, pImage)
        img_in: pImage = img._image
        while img_in.width > 128 and img_in.height > 128:
            img_in = img_in.resize((img_in.width // 2, img_in.height // 2))

        # get color histogram
        color_histogram: typing.Dict[str, int] = {}
        for x in range(0, img_in.width):
            for y in range(0, img_in.height):
                p: typing.Tuple[int, int, int] = img_in.getpixel((x, y))
                h: str = RGBColor(
                    Decimal(p[0] / 255.0), Decimal(p[1] / 255.0), Decimal(p[2] / 255.0)
                ).to_hex_string()
                color_histogram[h] = color_histogram.get(h, 0) + 1

        # compress color histogram
        sorted_histogram: typing.List[typing.Tuple[str, int]] = sorted(
            [(k, v) for k, v in color_histogram.items()],
            key=lambda x: x[1],
            reverse=True,
        )
        compressed_histogram: typing.List[typing.Tuple[str, int]] = []
        while len(sorted_histogram) > 0:
            c0, f0 = sorted_histogram[0]
            sorted_histogram.pop(0)
            similar_colors_and_frequencies = [
                t
                for t in sorted_histogram
                if ColorPaletteFromImage._dist(c0, t[0]) < 0.15 and t[1] < f0
            ]
            similar_colors = [t[0] for t in similar_colors_and_frequencies]
            compressed_histogram += [
                (c0, f0 + sum([t[1] for t in similar_colors_and_frequencies] + [0]))
            ]
            sorted_histogram = [
                t for t in sorted_histogram if t[0] not in similar_colors
            ]

        # return
        colors_out: typing.List[HexColor] = [
            HexColor(x[0]) for x in compressed_histogram
        ]
        return colors_out[:limit]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A license (or licence) is an official permission or permit to do, use, or own something
(as well as the document of that permission or permit). A license is granted by a party (licensor) to another party
(licensee) as an element of an agreement between those parties. In the case of a license issued by a government,
the license is obtained by applying for it. In the case of a private party, it is by a specific agreement,
usually in writing (such as a lease or other contract).

The simplest definition is "A license is a promise not to sue," because a license usually either permits
the licensed party to engage in an activity which is illegal, and subject to prosecution,
without the license (e.g. fishing, driving an automobile, or operating a broadcast radio or television station),
or it permits the licensed party to do something that would violate the rights of the licensing party
(e.g. make copies of a copyrighted work), which, without the license,
the licensed party could be sued, civilly, criminally, or both.
"""
import base64
import datetime
import json
import logging
import typing
from hashlib import sha256
from pathlib import Path

from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding

from borb.license.uuid import UUID

logger = logging.getLogger(__file__)


class License:
    """
    A license (or licence) is an official permission or permit to do, use, or own something
    (as well as the document of that permission or permit). A license is granted by a party (licensor) to another party
    (licensee) as an element of an agreement between those parties. In the case of a license issued by a government,
    the license is obtained by applying for it. In the case of a private party, it is by a specific agreement,
    usually in writing (such as a lease or other contract).
    """

    # fmt: off
    _LICENSE_COMPANY: typing.Optional[str] = None
    _LICENSE_USER_ID: typing.Optional[str] = None
    _LICENSE_VALID_FROM_IN_MS: typing.Optional[int] = None
    _LICENSE_VALID_UNTIL_IN_MS: typing.Optional[int] = None
    # fmt: on

    #
    # CONSTRUCTOR
    #

    #
    # PRIVATE
    #

    @staticmethod
    def _create(
        company: str = "",
        output_file: Path = Path(__file__).parent / "license_key.json",
        private_key_file: Path = Path("home/joris/borb_license_private_key.pem"),
        user_id: str = UUID.get(),
        valid_from_in_ms: int = int(datetime.datetime.now().timestamp() * 1000),
        valid_until_in_ms: int = int(datetime.datetime.now().timestamp() * 1000)
        + 7 * 24 * 60 * 60 * 1000,
    ) -> None:

        # set up dictionary
        # fmt: off
        license_dict: typing.Dict[str, str] = {
            "anonymous_user_id": user_id,
            "company": company,
            "valid_from_as_str": datetime.datetime.fromtimestamp(valid_from_in_ms / 1000).strftime("%Y-%m-%d %H:%M:%S"),
            "valid_until_as_str": datetime.datetime.fromtimestamp(valid_until_in_ms / 1000).strftime("%Y-%m-%d %H:%M:%S"),
        }
        # fmt: on

        # convert to bytes
        license_bytes: bytes = json.dumps(license_dict, indent=3).encode()
        license_bytes_unsigned_hash: bytes = sha256(license_bytes).hexdigest().encode()

        # Load private key from a file
        with open(private_key_file, "rb") as f:
            private_key = serialization.load_pem_private_key(f.read(), password=None)

        # Encode message with private key
        license_bytes_signed_hash = private_key.sign(
            license_bytes_unsigned_hash, padding.PKCS1v15(), hashes.SHA256()
        )

        # Base64 encode the encrypted message for transmission
        base64_license_bytes_signed_hash = base64.b64encode(license_bytes_signed_hash)

        # enter key in dictionary
        license_dict["license_key"] = base64_license_bytes_signed_hash.decode()

        # dump dictionary
        with open(output_file, "w") as fh:
            fh.write(json.dumps(license_dict, indent=3))

    @staticmethod
    def _datetime_str_to_ms(s: str) -> typing.Optional[int]:
        try:
            return int(
                datetime.datetime.strptime(s, "%Y-%m-%d %H:%M:%S").timestamp() * 1000
            )
        except:
            logger.warning("Unable to convert %s to a datetime" % str(s))
            return None

    #
    # PUBLIC
    #

    @staticmethod
    def get_company() -> typing.Optional[str]:
        """
        This function returns the company to which this License was granted
        :return:    the company to which this License was granted
        """
        return License._LICENSE_COMPANY

    @staticmethod
    def get_user_id() -> typing.Optional[str]:
        """
        This function returns the user_id to which this License was granted
        :return:    the user_id to which this License was granted
        """
        return License._LICENSE_USER_ID

    @staticmethod
    def get_valid_from_in_ms() -> typing.Optional[int]:
        """
        This function returns the start of the validity period of this License
        :return:    the start of the validity period of this License
        """
        return License._LICENSE_VALID_FROM_IN_MS

    @staticmethod
    def get_valid_until_in_ms() -> typing.Optional[int]:
        """
        This function returns the end of the validity period of this License
        :return:    the end of the validity period of this License
        """
        return License._LICENSE_VALID_UNTIL_IN_MS

    @staticmethod
    def register(path_to_license_file: Path) -> bool:
        """
        This function (attempts to) register a license (specified as a Path)
        :param path_to_license_file:    the license to register
        :return:                        True if the license is valid, False otherwise
        """

        # check whether the file exists
        if not path_to_license_file.exists():
            logger.warning("File %s does not exist" % str(path_to_license_file))
            return False

        # read license
        json_dict: typing.Optional[typing.Dict[str, str]] = None
        try:
            with open(path_to_license_file, "r") as fh:
                json_dict = json.loads(fh.read())
        except:
            # fmt: off
            logger.warning("Unable to convert contents of %s to JSON" % str(path_to_license_file))
            # fmt: on
            return False

        # remove key entry
        if json_dict is None:
            return False
        license_hash_0: bytes = json_dict.get("license_key", "").encode()
        json_dict.pop("license_key")

        # to bytes
        license_bytes1: bytes = json.dumps(json_dict, indent=3).encode()

        # hash
        license_hash_1: bytes = sha256(license_bytes1).hexdigest().encode()

        # Load public key(s) from a file
        key_dir: Path = Path(__file__).parent / "public_keys"
        assert key_dir.exists()
        assert key_dir.is_dir()
        for public_key_file in [key_dir / "public_key_001.pem"]:

            # read key
            with open(public_key_file, "rb") as f:
                public_key = serialization.load_pem_public_key(f.read())

            # Base64 decode the encoded message
            encrypted_message = base64.b64decode(license_hash_0)

            # Verify the signature using the public key
            try:
                public_key.verify(
                    encrypted_message,
                    license_hash_1,
                    padding.PKCS1v15(),
                    hashes.SHA256(),
                )
                # fmt: off
                License._LICENSE_COMPANY = json_dict.get("company", None)
                License._LICENSE_USER_ID = json_dict.get("anonymous_user_id", None)
                License._LICENSE_VALID_FROM_IN_MS = License._datetime_str_to_ms(json_dict.get("valid_from_as_str", None))
                License._LICENSE_VALID_UNTIL_IN_MS = License._datetime_str_to_ms(json_dict.get("valid_until_as_str", None))
                # fmt: on

                # check _LICENSE_VALID_FROM_IN_MS
                now_in_ms: int = int(datetime.datetime.now().timestamp() * 1000)
                if (
                    License._LICENSE_VALID_FROM_IN_MS is not None
                    and License._LICENSE_VALID_FROM_IN_MS > now_in_ms
                ):
                    # fmt: off
                    logger.warning("License %s is valid FROM %s" % (str(path_to_license_file), json_dict.get("valid_from_as_str", "N.A.")))
                    # fmt: on
                    License._LICENSE_COMPANY = None
                    License._LICENSE_USER_ID = None
                    License._LICENSE_VALID_FROM_IN_MS = None
                    License._LICENSE_VALID_UNTIL_IN_MS = None
                    return False

                # check _LICENSE_VALID_UNTIL_IN_MS
                if (
                    License._LICENSE_VALID_UNTIL_IN_MS is not None
                    and License._LICENSE_VALID_UNTIL_IN_MS < now_in_ms
                ):
                    # fmt: off
                    logger.warning("License %s is valid UNTIL %s" % (str(path_to_license_file), json_dict.get("valid_until_as_str", "N.A.")))
                    # fmt: on
                    License._LICENSE_COMPANY = None
                    License._LICENSE_USER_ID = None
                    License._LICENSE_VALID_FROM_IN_MS = None
                    License._LICENSE_VALID_UNTIL_IN_MS = None
                    return False

                return True
            except:
                continue

        # default
        return False
